from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

class IAM__Statement__Principal(Kwargs_To_Self):
    action         : str
    principal_type : str
    principal_value: str
    effect         : str

    def validate_data(self):
        return self.action and self.principal_type and self.principal_value and self.effect

    def statement(self):
        if self.validate_data():
            return   {'Action'   : self.action                                ,
                      'Effect'   : self.effect                                ,
                      'Principal': {self.principal_type: self.principal_value }}



from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

class Meta_Llama2(Kwargs_To_Self):
    model_id          : str
    prompt            : str
    temperature       : float    = 0.0
    top_p             : float    = 1.0
    max_gen_len       : int      = 1024

    def body(self):
        return dict(prompt        = self.prompt      ,
                    temperature   = self.temperature ,
                    top_p         = self.top_p       ,
                    max_gen_len   = self.max_gen_len )



from osbot_aws.AWS_Config import AWS_Config

from osbot_aws.apis.Session import Session
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self import cache_on_self


class Organizations(Kwargs_To_Self):
    aws_config: AWS_Config

    @cache_on_self
    def client(self):
        return Session().client('organizations')

    def account(self, account_id):
        return self.client().describe_account(AccountId=account_id)


    def accounts(self):
        return self.client().list_accounts()


import json

from osbot_aws.aws.iam.IAM import IAM
from osbot_utils.decorators.methods.cache import cache
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.decorators.methods.catch import catch

from osbot_aws.apis.Lambda import Lambda

from osbot_aws.apis.Session import Session


class Firehose:

    @cache_on_self
    def firehose(self):
        return Session().client('firehose')

    def _convert_to_bytes(self,data=''):
        if type(data) is not bytes:
            if type(data) is not str:
                data = json.dumps(data)
            data += '\n'                    # need to do this or we will get a massive list of entries
        return data.encode()

    @catch
    def add_processing_lambda(self, stream_name, lambda_name):
        function_arn = Lambda(lambda_name).function_arn()
        update_data =  { 'ProcessingConfiguration': {
                            'Enabled'   : True,
                            'Processors': [{ 'Type': 'Lambda',
                                              'Parameters': [{
                                                  'ParameterName' : 'LambdaArn',
                                                  'PameterValue': function_arn}]}]}}
        return self.destination_update(stream_name, update_data)

    def add_processing_lambda_set_lambda_permission(self, lambda_arn):
        statement_id = 'allow_firehose_to_invoke_lambda_function'
        action       = 'lambda:InvokeFunction'
        principal    = 'firehose.amazonaws.com'
        aws_lambda   = Lambda()
        aws_lambda.permission_delete(lambda_arn, statement_id)
        return aws_lambda.permission_add(lambda_arn, statement_id, action, principal, )

    def add_record(self, stream_name, record):
        params = {'DeliveryStreamName': stream_name ,
                  'Record': {'Data': self._convert_to_bytes(record) }}
        return self.firehose().put_record(**params)

    def add_records(self, stream_name, records):
        items = []
        for record in records:
            items.append({'Data': self._convert_to_bytes(record)})
        params = {'DeliveryStreamName': stream_name , 'Records': items}
        return self.firehose().put_record_batch(**params)

    @catch
    def destination_update(self, stream_name, update_data, destination_index=0):
        stream_info = self.stream(stream_name)
        version_id  = stream_info.get('VersionId')
        destination_id = stream_info.get('Destinations')[destination_index].get('DestinationId')

        params = { 'DeliveryStreamName'             : stream_name,
                   'CurrentDeliveryStreamVersionId' : version_id,
                   'DestinationId'                  : destination_id,
                   'ExtendedS3DestinationUpdate'    : update_data
                   }
        return self.firehose().update_destination(**params)



    @catch
    def stream(self, stream_name):
        return self.firehose().describe_delivery_stream(DeliveryStreamName=stream_name).get('DeliveryStreamDescription')

    @catch
    def stream_create(self, stream_name, s3_bucket, s3_prefix, compression='UNCOMPRESSED'):
        account_id = IAM().account_id()
        role       = 'firehose_delivery_role'
        params     = { "DeliveryStreamName": stream_name,
                       "DeliveryStreamType": "DirectPut",
                       'ExtendedS3DestinationConfiguration': {
                            "RoleARN"          : f"arn:aws:iam::{account_id}:role/{role}",
                            "BucketARN"        : f"arn:aws:s3:::{s3_bucket}",
                            "Prefix"           : f"{s3_prefix}/",
                            "CompressionFormat": compression   }}
        return self.firehose().create_delivery_stream(**params)

    @catch
    def stream_delete(self, stream_name):
        return self.firehose().delete_delivery_stream(DeliveryStreamName=stream_name)

    def streams(self):
        return self.firehose().list_delivery_streams().get('DeliveryStreamNames')





from osbot_aws.aws.apigateway.apigw_dydb.WS__Handle_Events import WS__Handle_Events

ws_handle_events = WS__Handle_Events()

def run(event, context):
    return ws_handle_events.handle_lambda(event, context)




from boto3.dynamodb.types import TypeSerializer, TypeDeserializer
from botocore.exceptions import ClientError

from osbot_aws.aws.dynamo_db.domains.DyDB__Table        import DyDB__Table
from osbot_aws.aws.dynamo_db.models.DyDB__Query__Builder import DyDB__Query__Builder
from osbot_utils.base_classes.Kwargs_To_Self            import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json                             import to_json_str
from osbot_utils.utils.Misc                             import list_set



class DyDB__Document(Kwargs_To_Self):
    table           : DyDB__Table
    document        : dict

    def __repr__(self):
        return f'[DyDB__Document]\n' + to_json_str(self.document)

    # support methods
    def db_query_builder(self):
        kwargs = dict(table_name = self.table.table_name,
                      key_name   = self.table.key_name,
                      key_value   = self.document_id())
        return DyDB__Query__Builder(**kwargs)

    def client(self):
        return self.table.dynamo_db.client()

    def delete(self):
        return self.table.delete_document(self.document_id())

    def document_id(self):
        return self.key_value()

    def field(self, field_name, reload_document=True):
        if reload_document:
            self.reload()
        return self.document.get(field_name)

    def fields(self, *field_names, reload_document=True):
        fields = {}
        if reload_document:
            self.reload()
        for field in field_names:
            fields[field] = self.document.get(field)
        return fields

    def fields_names(self):
        return list_set(self.document)

    def exp_key(self):
        return {self.key_name(): self.serialize_value(self.key_value()) }

    def key_name(self):
        return self.table.key_name

    def key_value(self):
        return self.document.get(self.key_name())

    def reload(self):
        document = self.table.document(self.document_id())
        if document:
            self.document = document
        return self.document

    def serialize_value(self, value):
        return self.type_serializer().serialize(value)

    def deserialize_value(self, value):
        return self.type_deserializer().deserialize(value)

    def value(self, field_name):
        return self.field(field_name, reload_document=False)

    def values(self):
        return self.document

    @cache_on_self
    def type_deserializer(self):
        return TypeDeserializer()

    @cache_on_self
    def type_serializer(self):
        return TypeSerializer()


    # action methods

    def add_field(self, field_name, field_value):
        return self.set_field(field_name=field_name, field_value=field_value)

    def add_to_list(self, list_field_name, new_list_element):
        kwargs   = self.db_query_builder().build__add_to_list(list_field_name, new_list_element)
        response = self.client().update_item(**kwargs)
        return response.get('Attributes')

    def add_to_set(self, field_name, element):
        kwargs   = self.db_query_builder().build__add_to_set(field_name, element)
        response = self.client().update_item(**kwargs)
        return response.get('Attributes')

    def dict_add_to_set(self, dict_field, field_name, element):
        kwargs = self.db_query_builder().build__dict_add_to_set(dict_field, field_name, element)
        response = self.client().update_item(**kwargs)
        return response.get('Attributes')

    def dict_delete_field(self, dict_field, field_key):
        kwargs = self.db_query_builder().build__dict_delete_field(dict_field, field_key)
        self.client().update_item(**kwargs)
        return self

    def dict_delete_from_set(self, dict_field, field_name, elements):
        kwargs = self.db_query_builder().build__delete_from_set(dict_field, field_name , elements)
        self.client().update_item(**kwargs)
        return self

    def dict_set_field(self, dict_field, field_key, field_value):
        kwargs = self.db_query_builder().build__dict_set_field(dict_field, field_key, field_value)
        self.client().update_item(**kwargs)
        return self

    def dict_dict_set_field(self, parent_dict_field, child_dict_field, field_key, field_value):
        kwargs = self.db_query_builder().build__dict_dict_set_field(parent_dict_field, child_dict_field, field_key, field_value)
        self.client().update_item(**kwargs)
        return self

    def dict_update_counter(self, dict_field, field_name, increment_by):
        kwargs = self.db_query_builder().build__dict_update_counter(dict_field, field_name, increment_by)
        self.client().update_item(**kwargs)
        return self

    def delete_field(self, field_name):
        if field_name not in self.fields_names():
            return False
        kwargs = self.db_query_builder().build__delete_field(field_name)
        self.client().update_item(**kwargs)
        return True

    def delete_item_from_list(self, list_field_name, item_index):
        kwargs = self.db_query_builder().build__delete_item_from_list(list_field_name, item_index)
        self.client().update_item(**kwargs)
        return self

    def delete_from_set(self, set_field_name, element):
        return self.delete_elements_from_set(set_field_name=set_field_name, elements={element})

    def delete_elements_from_set(self, set_field_name, elements):
        kwargs = self.db_query_builder().build__delete_elements_from_set(set_field_name, elements)
        self.client().update_item(**kwargs)
        return self

    def increment_field(self, field_name, increment_by):
        kwargs = self.db_query_builder().build__increment_field(field_name, increment_by)
        self.client().update_item(**kwargs)
        return self

    def increment_field__if_existing_value_bigger_than(self, field_name, increment_by, value: int):
        kwargs = self.db_query_builder().build__increment_field(field_name, increment_by)
        kwargs['ExpressionAttributeValues'][':zero'] = {'N': str(value)}
        kwargs['ConditionExpression'      ] = '#field_name > :zero'
        try:
            self.client().update_item(**kwargs)
            return True
        except ClientError as error:
            if error.response['Error']['Code'] == 'ConditionalCheckFailedException':
                return False
            else:
                raise

    def print(self):
        pprint(self.document)
        return self

    def print_field(self, field_name, new_line_before=True):
        value = self.field(field_name)
        if new_line_before:
            print()
        pprint(value)
        return value

    def reset_document(self):
        new_document  = {'id': self.document_id()}
        self.document =  self.table.add_document(new_document).get('document')
        return self

    def set_document(self,new_document):
        new_document[self.key_name()] = self.key_value()                             # make sure the id value is the same
        self.document = self.table.add_document(new_document).get('document')
        return self

    @remove_return_value(field_name='ResponseMetadata')
    def set_field(self, field_name, field_value):
        kwargs = self.db_query_builder().build__set_field(field_name, field_value)
        self.client().update_item(**kwargs)
        return self

    @remove_return_value(field_name='ResponseMetadata')
    def update_item(self, **kwargs):
        return self.client().update_item(**kwargs)


from osbot_utils.utils.Files import Files


def run(event, context):
    file_name = event.get('file_name')                                           # get file_name from lambda params
    tmp_path  = '/tmp'                                                           # location of lambda temp folder
    tmp_file  = Files.path_combine(tmp_path,file_name)                           # create file name (in temp folder)

    Files.write(tmp_file, 'some text')                                           # create file (with some text)

    return Files.find(tmp_path + '/*.*')                                         # return list of files in temp folder


from osbot_utils.utils.Dev import pprint

from osbot_aws.AWS_Config       import AWS_Config
from osbot_utils.utils.Files    import Files, file_create, path_combine, folder_create_temp, file_delete, folder_delete_recursively
from osbot_utils.utils.Misc     import random_string_and_numbers
from osbot_aws.aws.s3.S3        import S3
from osbot_utils.utils.Zip      import zip_folder


class Temp_S3_Zip_With_Lambda_File:
    def __init__(self, file_name=None,s3_bucket=None, s3_prefix=None, lambda_code=None, create_s3_file=True):
        self.aws_config     = AWS_Config()
        self.s3             = S3()
        self.create_s3_file = create_s3_file
        self.file_name      = file_name or f"temp_zip_file_{random_string_and_numbers()}"
        self.s3_bucket      = s3_bucket or self.aws_config.lambda_s3_bucket()
        self.s3_prefix      = s3_prefix or f'{AWS_Config().lambda_s3_folder_lambdas()}/unit_tests/temp_zips'
        self.s3_key         = f'{self.s3_prefix}/{self.file_name}.zip'
        self.s3_upload_ok   = False
        self.file_delete_ok = False
        self.lambda_code    = lambda_code or "def run(event, context): return 'from Temp_S3_Zip_With_Lambda_File, hello {0}'.format(event.get('name'))" # todo: use function_source_code(run)
        self.folder_temp    = None
        self.file_temp      = None
        self.file_zip       = None

    def __enter__(self):
        self.create_temp_file()
        self.s3_file_create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete_local_files()
        self.s3_file_delete()

#    def create_s3_file(self, new_code = None):
#

    def create_temp_file(self):
        self.folder_temp = folder_create_temp('tmp_lambda_')
        self.file_temp    = path_combine(self.folder_temp, f'{self.file_name}.py')
        file_create(self.file_temp, self.lambda_code)
        return self

    def delete_local_files(self):
        file_delete(self.file_zip)
        folder_delete_recursively(self.folder_temp)

    def s3_bucket_exists(self):
        return self.s3.bucket_exists(self.s3_bucket)

    def s3_file_create(self):
        temp_zip_file = self.zip()
        self.s3_upload_ok = self.s3.file_upload_to_key(temp_zip_file, self.s3_bucket, self.s3_key)

    def s3_file_delete(self):
        self.file_delete_ok = self.s3.file_delete(self.s3_bucket, self.s3_key)

    def s3_file_exists(self):
        return self.s3.file_exists(self.s3_bucket, self.s3_key)

    # def s3_file_contents(self):
    #     from gzip import GzipFile
    #     bytes = self.s3.file_bytes(self.s3_bucket, self.s3_key)
    #     return GzipFile(None, 'rb', fileobj=bytes)

    def s3_files_in_folder(self):
        return self.s3.find_files(self.s3_bucket, self.s3_prefix)

    def zip(self):
        self.file_zip = zip_folder(self.folder_temp)
        return self.file_zip

    def zip_bytes(self):
        with open(self.zip(), 'rb') as file_data:
            return file_data.read()


import uuid
from boto3.dynamodb.types                               import TypeDeserializer, TypeSerializer
from osbot_aws.apis.Session                             import Session
from osbot_utils.base_classes.Type_Safe                 import Type_Safe
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_utils.decorators.methods.remove_return_value import remove_return_value

DEFAULT_DOCUMENTS_MAX_ITEMS_TO_FETCH = 100000
DEFAULT_DOCUMENTS_FETCH_BATCH_SIZE   = 20000

class Dynamo_DB(Type_Safe):
    region_name : str = None
    endpoint_url: str = None
    # def __enter__(self                           ): return self
    # def __exit__ (self, exc_type, exc_val, exc_tb): pass

    # helpers
    @cache_on_self
    def client(self):
        return Session().client(service_name='dynamodb', region_name=self.region_name, endpoint_url=self.endpoint_url)

    @cache_on_self
    def client__dynamo_streams(self):
        return Session().client('dynamodbstreams')

    # main methods

    def document(self, table_name, key_name, key_value):
        key    = {key_name: {'S': key_value}}
        result = self.client().get_item(TableName=table_name, Key=key)
        item   = result.get('Item')
        return self.document_deserialize(item)

    def document_add(self, table_name, document):
        document_as_item = self.document_serialize(document)
        self.client().put_item(TableName=table_name, Item=document_as_item)
        result = dict(document=document, document_as_item=document_as_item)
        return result

    def document_delete(self, table_name, key_name, key_value):
        key = { key_name: {'S': key_value} }
        self.client().delete_item( TableName=table_name, Key=key ) # note: there is no clue from DynamoDB if it worked or not
        return True                                                #       so unless there was an exception thrown, assume it did

    def document_deserialize(self, item):
        if item:
            deserializer = TypeDeserializer()
            return {k: deserializer.deserialize(v) for k, v in item.items()}
        return {}

    @remove_return_value('ResponseMetadata')
    def document_update(self, table_name, key_name, key_value, update_data):
        # Initialize TypeSerializer to convert Python types to DynamoDB types
        serializer = TypeSerializer()

        # Construct the key for the item to update
        key = {key_name: {'S': key_value}}

        # Initialize the update expression components
        update_expression = "SET "
        expression_attribute_values = {}
        expression_attribute_names = {}

        for attribute_name, new_value in update_data.items():
            # Serialize the value to DynamoDB format
            dynamodb_value = serializer.serialize(new_value)

            # Create placeholders for attribute names and values
            attr_placeholder = f"#{attribute_name.replace('-', '_')}"
            val_placeholder = f":val_{attribute_name.replace('-', '_')}"

            # Add to the update expression, attribute names, and values
            update_expression += f"{attr_placeholder} = {val_placeholder}, "
            expression_attribute_names[attr_placeholder] = attribute_name
            expression_attribute_values[val_placeholder] = dynamodb_value

        # Remove the trailing comma and space from the update expression
        update_expression = update_expression.rstrip(', ')

        # Perform the update operation
        response = self.client().update_item(
            TableName=table_name,
            Key=key,
            UpdateExpression=update_expression,
            ExpressionAttributeNames=expression_attribute_names,
            ExpressionAttributeValues=expression_attribute_values,
            ReturnValues="UPDATED_NEW"
        )
        return response

    def document_serialize(self, document):
        serializer = TypeSerializer()
        return {k: serializer.serialize(v) for k, v in document.items()}

    def documents_add(self, table_name, documents):
        chunks = [documents[x:x + 25] for x in range(0, len(documents), 25)]        # Split the items list into chunks of 25 (DynamoDB batch write limit)
        responses = []

        for chunk in chunks:
            request_items = { table_name: [ {'PutRequest': {'Item': self.document_serialize(document)}}
                                              for document in chunk ] }
            response = self.client().batch_write_item(RequestItems=request_items)
            del response['ResponseMetadata']
            responses.append(response)
        result = dict(documents=documents, responses=responses)
        return result                                               # Contains unprocessed items

    def documents(self, table_name, key_name, keys_values):
        keys = []
        for key_value in keys_values:
            key = {key_name: {'S': key_value}}
            keys.append(key)

        chunks = [keys[x:x + 100] for x in range(0, len(keys), 100)]  # Split the items list into chunks of 100 (DynamoDB batch_get_item limit)

        all_responses        = []
        all_responses_raw    = []
        all_unprocessed_keys = []
        for chunk in chunks:
            kwargs           = {'RequestItems': {table_name: {'Keys': chunk}}}
            result           = self.client().batch_get_item(**kwargs)
            responses        = result.get('Responses')
            unprocessed_keys = result.get('UnprocessedKeys')
            all_responses_raw   .extend(responses.get(table_name))
            all_unprocessed_keys.append(unprocessed_keys)
        for response_raw in all_responses_raw:
            response = self.document_deserialize(response_raw)
            all_responses.append(response)

        return dict(all_responses        = all_responses        ,
                    all_unprocessed_keys = all_unprocessed_keys )

    def documents_all(self, table_name, batch_size=DEFAULT_DOCUMENTS_FETCH_BATCH_SIZE, max_fetch=DEFAULT_DOCUMENTS_MAX_ITEMS_TO_FETCH):
        #todo: figure out why this is only returning 64 records at the time
        items = []
        last_evaluated_key = None

        while True:
            scan_kwargs = {
                'TableName': table_name,
                'Limit'    : batch_size   # Specify the limit here
            }
            if last_evaluated_key:
                scan_kwargs['ExclusiveStartKey'] = last_evaluated_key

            response = self.client().scan(**scan_kwargs)
            items.extend(response.get('Items', []))
            last_evaluated_key = response.get('LastEvaluatedKey')

            if not last_evaluated_key:
                break
            if len(items) > max_fetch:
                break

        return [self.document_deserialize(item) for item in items]

    # todo:  figure out why code below returns only 64
    def documents_count(self, table_name):
        count_total  = 0
        response     = self.client().scan(TableName=table_name, Select='COUNT')
        count_total += response.get('Count', 0)
        while 'LastEvaluatedKey' in response:
            response = self.client().scan(TableName=table_name, Select='COUNT', ExclusiveStartKey=response['LastEvaluatedKey'])
            count_total += response.get('Count', 0)
        return count_total

    def documents_delete(self, table_name, key_name, keys_values):

        keys = [{key_name: {'S': key_value}} for key_value in keys_values]

        chunks = [keys[x:x+25] for x in range(0, len(keys), 25)]        # Split the keys list into chunks of 25 (DynamoDB batch write limit)
        responses = []

        for chunk in chunks:
            request_items = { table_name: [{'DeleteRequest': {'Key': key}} for key in chunk ] }
            response = self.client().batch_write_item(RequestItems=request_items)
            del response['ResponseMetadata']
            responses.append(response)

        return responses

    def documents_delete_all(self, table_name, key_name):
        all_keys      = self.documents_keys  (table_name=table_name, key_name=key_name)
        delete_result = self.documents_delete(table_name=table_name, key_name=key_name, keys_values=all_keys)
        delete_status = True
        for response in delete_result:
            if response.get('UnprocessedItems'):
                delete_status = False
                break
        return dict(deleted_keys=all_keys, delete_result=delete_result, delete_status=delete_status)

    def documents_ids(self, table_name, key_name, batch_size=DEFAULT_DOCUMENTS_FETCH_BATCH_SIZE, max_fetch=DEFAULT_DOCUMENTS_MAX_ITEMS_TO_FETCH):

        scan_kwargs = dict(ProjectionExpression = key_name  ,
                           Limit                = batch_size,
                           TableName            = table_name)
        response = self.client().scan(**scan_kwargs)

        all_items = []
        all_items.extend(response['Items'])
        while 'LastEvaluatedKey' in response and len(all_items) < max_fetch:
            scan_kwargs['ExclusiveStartKey'] = response.get('LastEvaluatedKey')
            response = self.client().scan(**scan_kwargs)
            all_items.extend(response['Items'])

        ids = []
        for item in all_items:
            ids.append(item.get(key_name).get('S'))
        return ids

    def documents_keys(self, table_name, key_name, key_type='S'):
        """
        Retrieves only the key attributes for all items in the DynamoDB table.
        :param table_name: Name of the DynamoDB table.
        :param key_name: The name of the primary key attribute.
        :return: A list of key values for all items in the table.
        """
        #response = self.client().scan(TableName=table_name, ProjectionExpression=key_name)
        #items    = response.get('Items', [])
        response = self.client().scan(TableName                = table_name      , # target table
                                      ProjectionExpression     = '#k'            , # Use a placeholder in the expression
                                      ExpressionAttributeNames = {'#k': key_name}) # Define the placeholder
        items = response.get('Items', [])
        return [item[key_name][key_type] for item in items]


    def table_create(self, table_name, key_name, with_streams=False):
        if self.table_exists(table_name):
            return False
        keySchema             = [{'AttributeName'    : key_name        , 'KeyType'           : 'HASH'}]
        attributeDefinitions  = [{'AttributeName'    : key_name        , 'AttributeType'     : 'S'}]
        #provisionedThroughput = { 'ReadCapacityUnits' : 5          , 'WriteCapacityUnits': 5      }
        kwargs   = { 'TableName'            : table_name           ,
                     'KeySchema'            : keySchema            ,
                     'AttributeDefinitions' : attributeDefinitions ,
                     'BillingMode'          : 'PAY_PER_REQUEST'    }
                     #'ProvisionedThroughput': provisionedThroughput }      # todo: add support for handling provisioned throughput
        if with_streams:
            kwargs['StreamSpecification'] = {'StreamEnabled': True, 'StreamViewType': 'NEW_IMAGE' }
        self.client().create_table(**kwargs)

        self.wait_for(waiter_name='table_exists', table_name=table_name)
        return True

    def wait_for(self, waiter_name, table_name, delay=1, max_attempts= 50):
        kwargs = dict(TableName    = table_name,
                      WaiterConfig = dict(Delay=delay, MaxAttempts=max_attempts))
        self.client().get_waiter(waiter_name).wait(**kwargs)

    def wait_for_table_exists(self, table_name):
        self.wait_for(waiter_name='table_exists', table_name=table_name)

    def table_delete(self, table_name, wait_for_deletion=True):
        if self.table_exists(table_name) is False:
            return False
        self.client().delete_table(TableName = table_name)
        if wait_for_deletion:
            self.wait_for(waiter_name='table_not_exists', table_name=table_name)
        return True

    def table_exists(self, table_name):
        return self.table_info(table_name) != {}

    def table_info(self, table_name):
        try:
            return self.client().describe_table(TableName=table_name).get('Table')
        except:
            return {}

    def table_status(self, table_name):
        return self.table_info(table_name).get('TableStatus')

    @remove_return_value(field_name='ResponseMetadata')
    def table_update(self, table_name, attribute_definitions=None, gsi_updates=None, stream_specification=None):
        update_kwargs = dict(TableName= table_name)
        if attribute_definitions:
            update_kwargs['AttributeDefinitions'       ] = attribute_definitions
        if gsi_updates:
            update_kwargs['GlobalSecondaryIndexUpdates'] = gsi_updates
        if stream_specification:
            update_kwargs['StreamSpecification'        ] = stream_specification

        return self.client().update_table(**update_kwargs )

    def tables(self):
        result = self.client().list_tables() or {}
        return result.get('TableNames') or []

    def streams(self):
        return self.client__dynamo_streams().list_streams().get('Streams')

    def random_id(self):
        return str(uuid.uuid4())


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Png import save_png_base64_to_file


class Stability_Stable_Diffusion(Kwargs_To_Self):
    model_id          : str
    text_prompts      : list
    cfg_scale         : int      = 10
    steps             : int      = 50
    seed              : int      = 42

    def set_text_prompts(self, prompts):
        if prompts is not list:
            prompts = [prompts]
        text_prompts = []
        for prompt in prompts:
            text_prompts.append( {"text": prompt})
        self.text_prompts = text_prompts

    def save_png_data(self, png_data, target_file=None):
        if target_file is None:
            target_file = "/tmp/bedrock_image.png"
        save_png_base64_to_file(png_data, target_file)


from osbot_aws.aws.s3.S3                import S3
from osbot_aws.aws.s3.S3__Minio         import S3__Minio
from osbot_aws.testing.TestCase__Dynamo_DB__Local import TestCase__Dynamo_DB__Local
from osbot_aws.testing.TestCase__Minio import TestCase__Minio
from osbot_utils.testing.Temp_Env_Vars  import Temp_Env_Vars
from unittest                           import TestCase
from osbot_utils.testing.Hook_Method    import Hook_Method
from osbot_utils.utils.Env              import get_env

TEST__AWS_ACCOUNT_ID     = '111122223333'
TEST__AWS_DEFAULT_REGION = 'eu-west-1'


class TestCase__Minio__Dynamo_DB(TestCase__Dynamo_DB__Local):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.test_case_minio = TestCase__Minio()
        cls.test_case_minio.setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.test_case_minio.tearDownClass()





from osbot_utils.utils.Misc import random_string

from osbot_aws.aws.ecs.ECS import ECS


class ECS_Cluster:
    def __init__(self, cluster_name=None):
        self.ecs          = ECS()
        self.cluster_name = cluster_name or random_string(prefix='test_ecs_cluster')


from osbot_aws.aws.bedrock.cache.Bedrock__Cache import Bedrock__Cache
from osbot_aws.aws.bedrock.cache.html.Bedrock_Cache__Html import Bedrock_Cache__Html
from osbot_utils.helpers.html.Tag__Base import Tag__Base
from osbot_utils.helpers.html.Tag__Div import Tag__Div
from osbot_utils.helpers.html.Tag__H import Tag__H
from osbot_utils.helpers.html.Tag__HR import Tag__HR
from osbot_utils.helpers.html.Tag__Head import Tag__Head
from osbot_utils.helpers.html.Tag__Html import Tag__Html
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Misc import list_set


class Bedrock_Cache__Html_With_Images(Bedrock_Cache__Html):


    def __init__(self):
        super().__init__()
        self.title     = 'AWS Bedrock Cached images'
        self.target_file = '/tmp/tmp-bedrock-images.html'


    def create__amazon_titan(self):
        div_images = []
        model_id  = 'amazon.titan-image-generator-v1'
        self.sub_title = f'for {model_id}'
        for request_data in self.cache.requests_data__with_model_id(model_id):
            request_hash        = request_data.get('_hash'    )
            request_id          = request_data.get('_id'      )
            request_comments    = request_data.get('_comments')
            request_body        = request_data.get('body'     )
            task_type           = request_body.get('taskType' )
            if task_type == 'TEXT_IMAGE':
                image_params = request_body.get('textToImageParams', {})
            elif task_type == 'IMAGE_VARIATION':
                image_params = request_body.get('imageVariationParams', {})
            elif task_type == 'INPAINTING':
                image_params = request_body.get('inPaintingParams', {})
            elif task_type == 'OUTPAINTING':
                image_params = request_body.get('outPaintingParams', {})
            else:
                image_params = {}
            image_text          = image_params.get('text')
            image_negative_text = image_params.get('negativeText', "")
            response_data       = self.cache.response_data_for__request_hash(request_hash)
            images              = response_data.get('images')
            image_base64        = images[0]

            div_image = Tag__Div(tag_classes = ['col', 'col-md-3', 'text-center'])
            div_badge          = Tag__Div(tag_classes=['badge','bg-primary'], inner_html = f"{task_type} #{request_id}")
            hr                 = Tag__HR()
            div_text_title     = Tag__Div(tag_classes=['var_name'], inner_html="Text:"          )
            div_text_value     = Tag__Div(inner_html=f"<strong>{image_text}</strong>"           )
            div_negative_title = Tag__Div(tag_classes=['var_name'], inner_html="Negative Text:" )
            div_negative_value = Tag__Div(inner_html=f"<strong>{image_negative_text}</strong>"  )
            div_comments_title = Tag__Div(tag_classes=['var_name'      ], inner_html="Comments:"      )
            div_comments_value = Tag__Div(tag_classes=['comments_value'], inner_html=request_comments  )
            img                = Tag__Base(tag_name="img", tag_classes=['base64-image','img-fluid'])
            img.attributes['src'] = f"data:image/jpeg;base64,{image_base64}"
            div_image.append(div_badge, hr,
                             img,
                             div_text_title   , div_text_value     ,
                             div_negative_title, div_negative_value,
                             div_comments_title, div_comments_value)
            div_images.append(div_image)

        self.row_elements = div_images

        self.create_html_file()

    def create__stable_diffusion(self):

        div_images = []
        #model_id = 'stability.stable-diffusion-xl-v0'
        model_id = 'stability.stable-diffusion-xl-v1'
        self.sub_title = f'for {model_id}'

        for request_data in self.cache.requests_data__with_model_id(model_id):
            request_hash        = request_data.get('_hash'    )
            request_id          = request_data.get('_id'      )
            #request_comments    = request_data.get('_comments')
            request_body        = request_data.get('body'     )
            response_data       = self.cache.response_data_for__request_hash(request_hash)
            artifacts           = response_data.get('artifacts')
            artifact            = artifacts[0]
            image_base64        = artifact['base64']
            cfg_scale           = request_body.get('cfg_scale')
            steps               = request_body.get('steps')
            text_prompts        = request_body.get('text_prompts')
            text_prompt         = text_prompts[0].get('text')
            image_text          = '....'
            div_image = Tag__Div(tag_classes = ['col', 'col-md-3', 'text-center'])
            div_badge          = Tag__Div(tag_classes=['badge','bg-primary'], inner_html = f"quality {steps} prompt #{cfg_scale}")
            hr                 = Tag__HR()
            div_text_title     = Tag__Div(tag_classes=['var_name'], inner_html="Text:"          )
            div_text_value     = Tag__Div(inner_html=f"<strong>{text_prompt}</strong>"           )
            # div_negative_title = Tag__Div(tag_classes=['var_name'], inner_html="Negative Text:" )
            # div_negative_value = Tag__Div(inner_html=f"<strong>{image_negative_text}</strong>"  )
            # div_comments_title = Tag__Div(tag_classes=['var_name'      ], inner_html="Comments:"      )
            # div_comments_value = Tag__Div(tag_classes=['comments_value'], inner_html=request_comments  )
            img                = Tag__Base(tag_name="img", tag_classes=['base64-image','img-fluid'])
            img.attributes['src'] = f"data:image/jpeg;base64,{image_base64}"
            div_image.append(div_badge, hr,
                             img,
                             div_text_title   , div_text_value     )
                             # div_negative_title, div_negative_value,
                             # div_comments_title, div_comments_value)
            div_images.append(div_image)

            # print(cfg_scale, steps)
            # print(len(image_base64))
        self.row_elements = div_images
        self.create_html_file()




from osbot_aws.aws.ec2.EC2 import EC2
from osbot_aws.aws.ec2.EC2_Instance import EC2_Instance
from osbot_aws.aws.ec2.utils.EC2__Create__Instance import EC2__Create__Instance
from osbot_utils.base_classes.Type_Safe import Type_Safe


class EC2__Temp__Instance(Type_Safe):
    ec2_create_instance : EC2__Create__Instance  = None
    ec2_instance        : EC2_Instance           = None
    ec2_instance_id     : str
    wait_for_ssh        : bool                   = True
    ec2                 : EC2

    def __enter__(self):
        self.ec2_create_instance = EC2__Create__Instance(ec2=self.ec2)
        self.ec2_instance_id     = self.ec2_create_instance.create_instance__with_amazon_linux__t3_nano__ssh_sg__spot()
        self.ec2_instance        = EC2_Instance(instance_id=self.ec2_instance_id, ec2=self.ec2)
        if self.wait_for_ssh:
            self.ec2_instance.wait_for_ssh()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ec2_instance.delete()
        return False

    def ssh(self):
        return self.ec2_instance.ssh(self.ec2_create_instance.path_key_file(), 'ec2-user')




from osbot_utils.decorators.lists.index_by              import index_by
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_utils.decorators.methods.required_fields     import required_fields
from osbot_aws.AWS_Config                               import AWS_Config
from osbot_aws.aws.s3.S3                                import S3
from osbot_aws.apis.Session                             import Session
from osbot_utils.testing.Temp_Folder                    import Temp_Folder
from osbot_utils.testing.Temp_Zip                       import Temp_Zip
from osbot_utils.utils.Files                            import file_bytes, temp_folder, temp_file, file_name, path_combine, folder_copy
from osbot_utils.utils.Http                             import GET_bytes_to_file
from osbot_utils.utils.Process                          import run_process
from osbot_utils.utils.Zip                              import folder_zip


class Lambda_Layer:
    def __init__(self, layer_name='', runtimes=None, license_info=None, s3_bucket=None, s3_folder=None, description=None, region=None, account_id=None):
        self.aws_config      = AWS_Config()
        self.layer_name      = layer_name.replace('.', '-')
        self.runtimes        = runtimes         or ['python3.11', 'python3.10', 'python3.9']# 'python3.8', 'python3.7', 'python3.6']
        self.license_info    = license_info     or 'https://github.com/owasp-sbot/OSBot-AWS/blob/master/LICENSE'
        self.description     = description      or ''
        self.s3_bucket       = s3_bucket        or self.aws_config.lambda_s3_bucket()
        self.s3_folder       = s3_folder        or self.aws_config.lambda_s3_folder_layers()
        self.region          = region           or self.aws_config.aws_session_region_name()
        self.account_id      = account_id       or self.aws_config.aws_session_account_id()
        self.s3_key          = f'{self.s3_folder}/{self.layer_name}.zip'
        #self.version_number  = version_number       # todo fix the use of this value


    # cached dependencies

    def arn(self, version):
        return f'arn:aws:lambda:{self.region}:{self.account_id}:layer:{self.layer_name}:{version}'

    def arn_latest(self):
        version = self.latest_version()
        if version:
            return self.arn(version)

    @cache_on_self
    def client(self):
        return Session().client('lambda')

    def content(self, version=None):
        return self.info(version).get('Content')

    @remove_return_value('ResponseMetadata')
    def version_by_arn(self):
        kwargs = {'Arn': self.arn_latest()}
        return self.client().get_layer_version_by_arn(**kwargs)

    @cache_on_self
    def s3(self):
        return S3()

    def create_from_folder(self, folder):
        zip_file  = folder_zip(folder)
        zip_bytes = file_bytes(zip_file)
        return self.create_from_zip_bytes(zip_bytes)

    # def create_from_folder_via_s3(self, folder):
    #     zip_file = folder_zip(folder)
    #     return self.create_from_zip_file_via_s3(zip_file)

    def create_from_folder_via_s3(self, source_folder, ignore_pattern=None):
        if ignore_pattern is None:
            ignore_pattern = ['*.pyc', '.DS_Store', '__pycache__', '.env']

        with Temp_Folder(temp_prefix='lambda_layer') as _:
            destination = path_combine(_.path(), 'python')
            folder_copy(source=source_folder, destination=destination, ignore_pattern=ignore_pattern)
            with Temp_Zip(_.path()) as zip_file:
                result = self.create_from_zip_file_via_s3(zip_file.path())
                return  result.get('LayerVersionArn')

    def create_from_pip(self, package_name, pip_executable='pip3'):
        path_install = temp_folder()
        install_result = run_process(pip_executable, ['install','-t',path_install,package_name])
        if install_result.get('error') is None and install_result.get('stderr').startswith('ERROR') is False:
            return self.create_from_folder(path_install)
        else:
            return {'status': 'error', 'error':install_result.get('stderr'), 'stdout': install_result.get('stdout')}

    def create_from_zip_file_via_s3(self, zip_file):
        self.upload_layer_zip_file_to_s3(zip_file)
        return self.create_from_s3()

    @remove_return_value('ResponseMetadata')
    @required_fields(['layer_name'])
    def create_from_zip_bytes(self, zip_bytes):
        params = { 'LayerName'         : self.layer_name           ,
                   'Description'       : self.description          ,
                   'Content'           : {  'ZipFile' : zip_bytes },
                   'LicenseInfo'       : self.license_info         ,
                   'CompatibleRuntimes': self.runtimes             }
        return self.client().publish_layer_version(**params)

    @remove_return_value('ResponseMetadata')
    @required_fields(['layer_name'])
    def create_from_s3(self):
        params = {'LayerName'           : self.layer_name               ,
                    'Description'       : self.description              ,
                    'Content'           : { 'S3Bucket': self.s3_bucket  ,
                                            'S3Key'   : self.s3_key    },
                    'CompatibleRuntimes': self.runtimes                 ,
                    'LicenseInfo'       : self.license_info
                  }
        return self.client().publish_layer_version(**params)

    def upload_layer_zip_file_to_s3(self, zip_file):
        return self.s3().file_upload_to_key(zip_file, self.s3_bucket, self.s3_key)

    def code_zip(self):
        version_arn = self.latest().get('LayerVersionArn')
        if version_arn:
            url_code = self.client().get_layer_version_by_arn(Arn=version_arn).get('Content',{}).get('Location')
            if url_code:
                path_target = temp_file(extension='.zip')
                return GET_bytes_to_file(url_code, path_target)

    def delete(self):
        for version_info in self.versions():
            self.delete_version(version_info.get('Version'))

        return self.exists() is False

    def delete_version(self, version):
        return self.client().delete_layer_version(LayerName=self.layer_name, VersionNumber=version)

    def exists(self):
        return self.latest() != {}


    @remove_return_value('ResponseMetadata')
    def info(self, version=None):
        return self.version(version)

    #def info_by_arn(self, version=None):

    def info_by_arn(self):
       return self.version_by_arn()

    def latest_version(self):
        return self.latest().get('Version')

    def latest(self):
        results = self.versions(max_items=1)
        if len(results) > 0:
            return results[0]
        return {}

    @index_by
    def layers(self):
        return self.client().list_layers().get('Layers')

    def policy(self, version=None):
        kwargs = {'LayerName'    : self.layer_name,
                  'VersionNumber': version or self.latest_version()}
        return self.client().get_layer_version_policy(**kwargs)  # return self.latest()


    def s3_folder_files(self):
        return self.s3().find_files(self.s3_bucket, self.s3_folder)

    def version(self, version=None):
        kwargs = {'LayerName': self.layer_name,
                  'VersionNumber': version or self.latest_version()}
        return self.client().get_layer_version(**kwargs)

    def versions(self, max_items=None):
        kwargs = {'LayerName': self.layer_name}
        if max_items:
            kwargs['MaxItems'] = max_items
        return self.client().list_layer_versions(**kwargs).get('LayerVersions')


from functools                                  import wraps
from botocore.client                            import BaseClient
from osbot_utils.helpers.Print_Table            import Print_Table
from osbot_utils.utils.Dev                      import pformat
from osbot_utils.base_classes.Kwargs_To_Self    import Kwargs_To_Self
from osbot_utils.testing.Duration               import Duration
from osbot_utils.testing.Hook_Method            import Hook_Method
from osbot_utils.utils.Objects                  import obj_data


# todo: create unit tests specifically for this class
# decorator

# todo refactor ctor of print_boto3_calls to make these options available (and easier to use)
# config__print_args          : bool = True
# config__print_calls         : bool = True
# config__print_call_stack    : bool = False
# config__print_return_value  : bool = True
# config__print_pformat_args  : bool = True
# see above the multipe options that can be set via the decorator

def print_boto3_calls(show=True, show_args=True, show_calls= False, show_return=True, **decorator_kwargs):
    decorator_kwargs['config__print_args'        ] = show_args
    decorator_kwargs['config__print_call_stack'  ] = show_calls
    decorator_kwargs['config__print_calls'       ] = show
    decorator_kwargs['config__print_return_value'] = show_return
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            with View_Boto3_Rest_Calls(**decorator_kwargs):         # Pass decorator_kwargs to the View_Boto3_Rest_Calls class
                return function(*args, **kwargs)
        return wrapper
    return decorator


class View_Boto3_Rest_Calls(Kwargs_To_Self):
    config__print_args          : bool = True
    config__print_calls         : bool = True
    config__print_call_stack    : bool = False
    config__print_return_value  : bool = True                       # todo: add better support for this (for example not showing the column if this is False)
    config__print_pformat_args  : bool = True
    target_module                      = BaseClient
    target_method               : str  = "_make_api_call"
    hook_method                        = None
    total_duration              : Duration

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.total_duration.print_result = False

    def __enter__(self):
        self.hook_method = Hook_Method(target_module=self.target_module, target_method=self.target_method)
        self.total_duration.start()
        return self.hook_method.__enter__()

    def __exit__(self, exception_type, exception_value, traceback):
        self.total_duration.end()
        if self.config__print_calls:
            self.print_calls()
        return self.hook_method.__exit__(type=exception_type, value=exception_value,traceback=traceback)

    def print_data(self):
        total_duration  = round(self.total_duration.seconds(), 2)
        title           = f"BOTO3 REST calls (via BaseClient._make_api_call)"
        footer          = f"Total Duration: {total_duration:6} secs | Total calls: {len(self.hook_method.calls)} {' ':8}"
        #headers  = f"{'#':2} | {'Method':22} | {'Duration':7} | {'Params':42} | {'Return Value':42}"
        headers_data = { '#': 2, 'Method': 22, 'Duration': 7, 'Params': 42, 'Return Value': 42 }
        headers = " | ".join(f"{header:{width}}" for header, width in headers_data.items())
        rows_data = []
        for call in self.hook_method.calls:
            api_name     = call.get('args'  )[1]
            args         = call.get('args'  )[2:]
            call_stack   = call.get('call_stack')
            kwargs       = call.get('kwargs')           # see if this is relevant to any calls
            duration     = call.get('duration')
            exception    = call.get('exception')
            index        = call.get('index')
            return_value = call.get('return_value')
            if return_value and 'ResponseMetadata' in return_value:
                del return_value['ResponseMetadata']
            if exception:
                # exception_data = obj_data(exception, name_width=100, value_width=1000)
                # return_value_str = pformat(exception_data)
                return_value_str = pformat(exception.args)
            else:
                if self.config__print_return_value:
                    return_value_str = pformat(return_value)
                else:
                    return_value_str = '(hidden)'

            if self.config__print_args is False:
                args = '(hidden)'
            if self.config__print_pformat_args:
                args= pformat(args)

            row_data = dict(api_name     = api_name         ,
                            args         = args             ,
                            index        = index            ,
                            kwargs       = kwargs           ,
                            duration     = duration         ,
                            return_value = return_value_str )
            rows_data.append(row_data)
            if self.config__print_call_stack:               # todo: this needs improvement since at the momment the call stacks are show at the top of the console logs
                call_stack.print()

        return { 'headers'      : headers       ,
                 'headers_data' : headers_data  ,
                 'footer'       : footer        ,
                 'rows_data'    : rows_data     ,
                 'title'        : title         }

    def print_calls(self):
        print_data = self.print_data()
        rows_data  = print_data.get('rows_data')
        if rows_data:
            order = ['index', 'api_name', 'duration', 'args', 'kwargs', 'return_value']
        else:
            order = []
        with Print_Table() as _:
            _.set_title(print_data.get('title'))
            _.set_footer(print_data.get('footer'))
            _.add_data(rows_data)
            _.set_order(*order)
            _.print()


    def map_data(self):
        pass


from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.aws.ec2.EC2 import EC2
from osbot_aws.aws.sts.STS import STS
from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.helpers.Local_Cache import Local_Cache

AWS__AMIS__MAPPINGS = {'eu-west-1' : {'x86_64': { 'amazon-linux-3': 'ami-04fe22dfadec6f0b6' },
                                      'arm64' : { 'amazon-linux-3': 'ami-05172b510cbeb4f59' }},
                       'eu-west-2' : {'x86_64': { 'amazon-linux-3': 'ami-06373f703eb245f45' },
                                      'arm64' : { 'amazon-linux-3': 'ami-06c0324315b2eb40d' }}}



class AMI(Type_Safe):
    aws_config: AWS_Config
    ec2       : EC2
    # def __init__(self):
    #     self.current_region_name = self.sts().current_region_name()

    # @cache_on_self
    # def sts(self):
    #     return STS()
    #
    def all_amis__local_cache(self):
        return Local_Cache('all_amazon_amis')

    def all_amis__from_amazon__cached(self):                    # this is a 15Mb file!
        local_cache = self.all_amis__local_cache()
        if local_cache.cache_exists():
            return local_cache.data().get('amazon_amis')
        else:
            all_amazon_amis = self.ec2.amis(owner='amazon')
            amis_data = {'amazon_amis': all_amazon_amis}
            local_cache.add_data(amis_data)
            return all_amazon_amis

    @cache_on_self
    def ami_mappings_per_region(self):                                                          # todo find way to pre-create these in a performant way (for all regions)
        return AWS__AMIS__MAPPINGS

    def get_ami_for_image(self, image_name, region_name, architecture):
        return self.ami_mappings_per_region().get(region_name, {}).get(architecture, {}).get(image_name,'')

    def amazon_linux_3(self, architecture='x86_64'):
        image_name = 'amazon-linux-3'
        return self.get_ami_for_image(image_name=image_name, region_name=self.aws_config.region_name(), architecture=architecture)

    def amazon_linux_3_x86_64(self):
        return self.amazon_linux_3(architecture='x86_64')

    def amazon_linux_3_arm64(self):
        return self.amazon_linux_3(architecture='arm64')


from datetime import date, timedelta

from osbot_aws.aws.iam.IAM import IAM
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.Lambda import Lambda
from osbot_aws.apis.Session import Session


class API_Gateway:

    def __init__(self):
        pass

    @cache_on_self
    def api_gateway(self):
        return Session().client('apigateway')

    # helper methods
    def _call_method(self, method_name, params):
        try:
            return getattr(self.api_gateway(), method_name)(**params)
        except Exception as error:
            return {'error': f'{error}'}

    # see this tweet to understand the use of the 'data_key' param https://twitter.com/DinisCruz/status/1226113182240038912
    # see this tweet thread to see more info about performance issues with this api https://twitter.com/DinisCruz/status/1226504297439023104
    # todo: add support for getting all the data using the position field
    def _call_method_return_items(self, method_name, params={}, index_by='id', data_key='items'):
        try:
            raw_data = getattr(self.api_gateway(), method_name)(**params)
            data = {}
            for item in raw_data.get(data_key):
                data[item.get(index_by)] = item
            return data
        except Exception as error:
            return {'error' : f'{error}'}

    def _call_method_return_arrays(self, method_name, params={}, index_by='id', data_key='items'):
        try:
            raw_data = getattr(self.api_gateway(), method_name)(**params)
            data = {}

            for key,value in raw_data.get(data_key).items():
                data[key] = value
            return data
        except Exception as error:
            return {'error' : f'{error}'}

    def _get(self, target, params={}, index_by='id'):
        return self._call_method_return_items(method_name=f"get_{target}", params=params, index_by=index_by)

    def _get_using_api_id(self, target, api_id, index_by='id', data_key='items'):
        return self._call_method_return_items(method_name=f"get_{target}", params={'restApiId': api_id}, index_by=index_by, data_key=data_key)

    def _index_by(self, values, index_by=None):
        if index_by is None:
            return list(values)
        results = {}
        for item in values:
            results[item.get(index_by)] = item
        return results

    # main methods
    def account(self):
        return self.api_gateway().get_account()

    #@aws_retry()  #todo add the code below to the aws_retry method (this is needed by operations that are affected by delays in IAM deployments
    def account_update(self, operation, path, value, _from, retry_count=5, retry_wait=2):
#        while (retry_count > 0):
#            print(f"Retry #{retry_count}")
#            try:
                patchOperations=[{
                                    'op'   : operation,  #'add'|'remove'|'replace'|'move'|'copy'|'test',
                                 'path' : path     ,
                                 'value': value    ,
                                 'from' : _from
                              }]
                return self.api_gateway().update_account(patchOperations=patchOperations)
#            except Exception as error:
#                print(f"error: {error}")
#            retry_count -= 1

    def api_keys(self, index_by='id',include_values=False):
        return self._call_method_return_items(method_name="get_api_keys", params={'includeValues':include_values}, index_by=index_by)

    def api_key(self, api_key, include_value=False):
        try:
            return self.api_gateway().get_api_key(apiKey=api_key,includeValue=include_value)
        except:
            return None

    def api_key_create(self,key_name, enabled=True):
        return self.api_gateway().create_api_key(name=key_name,enabled=enabled)

    def api_key_delete(self, key_id_or_name):
        if self.api_key_exists(key_id_or_name):                                     # see if it an api_key value
            self.api_gateway().delete_api_key(apiKey=key_id_or_name)
            return True

        for key_id, value in self.api_keys().items():                           # try to find api_key value
            if value.get('name') == key_id_or_name:                             # via its name
                self.api_gateway().delete_api_key(apiKey=key_id)
                return True
        return False

    def api_key_exists(self, api_key):
        if self.api_key(api_key):
            return True
        return False

    def deployments(self, api_id):
        return self._call_method_return_items(method_name="get_deployments", params={'restApiId':api_id})

    def deployment_create(self, api_id, stage):
        params = {'restApiId': api_id,
                  'stageName': stage}
        return self._call_method('create_deployment',params)

    def domain_name_add_path_mapping(self, rest_api_id, domain_name, base_path, stage='prod'):
        try:
            params = { 'domainName' : domain_name ,
                       'basePath'   : base_path   ,
                       'restApiId'  : rest_api_id ,
                       'stage'      : stage       }
            return self.api_gateway().create_base_path_mapping(**params)
        except Exception as error:
            return {'error': f'{error}'}
    def domain_name_create__regional(self, domain_name, certificate_arn):
        try:
            params = { 'domainName'    : domain_name,
                       'regionalCertificateArn': certificate_arn,
                       'endpointConfiguration': { 'types': ['REGIONAL'] },
                       'securityPolicy'    :'TLS_1_2'}
            return self.api_gateway().create_domain_name(**params)
        except Exception as error:
            return {'error':  f'{error}'}

    def domain_name_delete(self, domain_name):
        try:
            return self.api_gateway().delete_domain_name(domainName=domain_name)
        except Exception as error:
            return {'error':  f'{error}'}

    def domain_names(self, index_by=None):
        return self._index_by(self.api_gateway().get_domain_names().get('items'), index_by=index_by)

    def integration(self, api_id, resource_id, http_method):
        return self.api_gateway().get_integration(restApiId=api_id, resourceId=resource_id, httpMethod=http_method)

    def integration_create__http(self, api_id, resource_id, uri, http_method='GET', integration_http_method='GET'):
        input_type        = 'HTTP'
        try:
            return self.api_gateway().put_integration(restApiId=api_id,
                                                    resourceId=resource_id,
                                                    httpMethod=http_method,
                                                    integrationHttpMethod=integration_http_method,
                                                    type=input_type,
                                                    uri=uri)
        except Exception as error:
            return f'{error}'

    def integration_create__lambda(self, api_id, resource_id, lambda_name, http_method):
        iam         = IAM()
        aws_acct_id = iam.account_id()
        aws_region  = iam.region()

        input_type = 'AWS_PROXY'
        uri = f'arn:aws:apigateway:{aws_region}:lambda:path/2015-03-31/functions/arn:aws:lambda:{aws_region}:{aws_acct_id}:function:{lambda_name}/invocations'
        integration_http_method = 'POST'
        try:
            return self.api_gateway().put_integration(restApiId=api_id, resourceId=resource_id, httpMethod=http_method,
                                                    integrationHttpMethod=integration_http_method,type=input_type, uri=uri)
        except Exception as error:
            return f'{error}'

    def integration_add_permission_to_lambda(self,api_id, lambda_name):
        # create permission to allow lambda function to be invoked by API Gateway
        iam           = IAM()
        aws_acct_id   = iam.account_id()
        aws_region    = iam.region()
        aws_lambda    = Lambda(lambda_name)
        function_arn  = aws_lambda.function_arn()#'gw_bot.lambdas.dev.hello_world'
        statement_id  = 'allow-api-gateway-invoke'
        action        = 'lambda:InvokeFunction'
        principal     = 'apigateway.amazonaws.com'
        source_arn    = f'arn:aws:execute-api:{aws_region}:{aws_acct_id}:{api_id}/*/GET/'

        aws_lambda.permission_delete(function_arn, statement_id) # remove in case there was already a permission with this name
        return aws_lambda.permission_add(function_arn, statement_id, action, principal, source_arn)

    def integration_response(self, api_id, resource_id, http_method, status_code):
        params = {'restApiId': api_id,
                  'resourceId': resource_id,
                  'httpMethod': http_method,
                  'statusCode': status_code,
                  }
        return self._call_method('get_integration_response', params)

    def integration_response_create(self, api_id, resource_id, http_method, status_code,response_templates):
        params = {'restApiId'        : api_id,
                  'resourceId'       : resource_id,
                  'httpMethod'       : http_method,
                  'statusCode'       : status_code,
                  'responseTemplates': response_templates
                  }
        return self._call_method('put_integration_response', params)

    def method(self, api_id, resource_id, http_method):
        try:
            return self.api_gateway().get_method(restApiId=api_id, resourceId= resource_id, httpMethod= http_method)
        except Exception as error:
            return {'error': f'{error}'}

    def method_create(self, api_id, resource_id, http_method, authorization_type='NONE'):
        return self._call_method("put_method", { 'restApiId': api_id, 'resourceId':resource_id, 'httpMethod':http_method, 'authorizationType':authorization_type})
#        return self.api_gateway().put_method(restApiId=api_id, resourceId=resource_id, httpMethod=http_method, authorizationType=authorization_type)

    def method_delete(self, api_id, resource_id, http_method):
        return self.api_gateway().delete_method(restApiId=api_id, resourceId=resource_id, httpMethod=http_method)

    def method_invoke_test(self, api_id,resource_id,http_method, path_with_query_string='',body='',headers={}):
        params = { 'restApiId'          : api_id ,
                   'resourceId'         : resource_id,
                   'httpMethod'         : http_method,
                   'pathWithQueryString': path_with_query_string,
                   'body'               : body   ,
                   'headers'            : headers}
        return self._call_method("test_invoke_method", params)

    def method_response(self,api_id,resource_id,http_method,status_code):
        params = { 'restApiId' : api_id ,
                   'resourceId': resource_id,
                   'httpMethod': http_method,
                   'statusCode': status_code,
                   }
        return self._call_method('get_method_response', params)

    def method_response_create(self,api_id,resource_id,http_method,status_code,response_models):
        params = { 'restApiId'     : api_id ,
                   'resourceId'    : resource_id ,
                   'httpMethod'    : http_method ,
                   'statusCode'    : status_code ,
                   'responseModels': response_models
                   }
        return self._call_method('put_method_response', params)

    def models(self, api_id):
        return self._get_using_api_id('models', api_id)


    def stage(self, api_id, stage_name):
        return self._call_method("get_stage", {'restApiId' : api_id, 'stageName':stage_name})

    def stage_url(self,api_id, region, stage, resource=''):
        return f'https://{api_id}.execute-api.{region}.amazonaws.com/{stage}/{resource}'

    def stages(self, api_id, index_by='deploymentId'):
       return self._get_using_api_id('stages', api_id, index_by=index_by, data_key='item')


    #def stages(self, api_id, index_by='deploymentId'):
    #    return self._call_method_return_items(method_name="get_stages", params={'restApiId':api_id},index_by=index_by)

    def resource(self, api_id, path):
        return self.resources(api_id, index_by='path').get(path,{})

    def resource_id(self, api_id, path):
        return self.resource(api_id,path).get('id')

    def resource_methods(self, api_id, path):
        return list(set(self.resource(api_id,path).get('resourceMethods',[])))

    def resource_create(self, api_id, parent_id, path):
        return self.api_gateway().create_resource(restApiId=api_id, parentId=parent_id,pathPart=path)

    def resource_delete(self, api_id, resource_id):
        return self.api_gateway().delete_resource(restApiId=api_id, resourceId=resource_id)

    def resources(self, api_id_or_name, index_by='id'):
        result = self._get_using_api_id('resources', api_id=api_id_or_name, index_by=index_by)
        if result.get('error') is None:
            return result
        rest_apis = self.rest_apis(index_by='name')
        if api_id_or_name in rest_apis:
            api_id = rest_apis.get(api_id_or_name).get('id')
            return self._get_using_api_id('resources', api_id=api_id, index_by=index_by)
        return {'error': f'API not found: {api_id_or_name}'}

    def rest_api(self, api_name):
        return self.rest_apis(index_by='name').get(api_name,{})

    def rest_api_create(self, api_name):
        rest_apis = self.rest_apis(index_by='name')                   # get existing Rest APIs
        if api_name in rest_apis:                                     # see if it already exists
            return rest_apis.get(api_name)                            # return if it does
        params = { 'name'                 : api_name               ,
                   'endpointConfiguration': {"types": ["REGIONAL"]}}
        return self.api_gateway().create_rest_api(**params)              # if not, create it

    def rest_api_delete(self, api_id):
        if self.rest_api_exists(api_id):
            self.api_gateway().delete_rest_api(restApiId=api_id)
        return self.rest_api_not_exists(api_id)

    def rest_api_exists(self, api_id):
        return api_id and self.rest_api_info(api_id).get('id') == api_id

    def rest_api_id(self, api_name):
        return self.rest_api(api_name).get('id')

    def rest_api_info(self, api_id):
        try:
            return self.api_gateway().get_rest_api(restApiId=api_id)
        except Exception as error:
            return {'error': f'{error}'}

    def rest_api_not_exists(self, api_id):
        return self.rest_api_exists(api_id) is False

    def rest_apis(self, index_by='id'):
        return self._call_method_return_items(method_name="get_rest_apis",index_by=index_by)

    def usage_raw(self, usage_plan_id, days):
        start_date = (date.today() - timedelta(days=days)).strftime("%Y-%m-%d")
        end_date   = date.today().strftime("%Y-%m-%d")
        raw_data   = self._call_method_return_arrays(method_name= 'get_usage',params={'usagePlanId':usage_plan_id ,'startDate'  : start_date   , 'endDate'    : end_date})
        return raw_data

    def usage(self, usage_plan_id, days):
        if days > 90:
            days =90
        results = {}
        raw_data = self.usage_raw(usage_plan_id, days)
        for key, value in raw_data.items():
            key_results = {}
            for i in range(0,days):
                row_date = (date.today() - timedelta(days=i)).strftime("%Y-%m-%d")
                key_results[row_date] = value[days-i][0]

            results[key] = key_results
        return results

    # this method helps to create a table that is usable by graph engines like Google Charts
    def usage__as_chart_data(self,usage_plan_id, days):
        api_keys = self.api_keys()

        usage = self.usage(usage_plan_id, days)
        headers = ['Days','Totals']
        keys = sorted(list(set(usage)))
        for key in keys:
            name = api_keys[key].get('name')
            headers.append(name)
        rows = [headers]

        if len(keys) > 0:
            days = usage[keys[0]].keys()
            for day in days:
                row_key = day.replace('2020-','')
                row = [row_key]
                total = 0
                for key in keys:
                    value = usage[key][day]
                    row.append(value)
                    total += value
                row.insert(1,total)
                rows.insert(1,row)
                
        return rows

    def usage_plan_keys(self, usage_plan_id):
        return self._get('usage_plan_keys', {'usagePlanId':usage_plan_id})


    def usage_plan_add_key(self,usage_plan_id, key_id):

         return self._call_method('create_usage_plan_key', { 'usagePlanId': usage_plan_id,
                                                             'keyId'      : key_id,
                                                             'keyType'    : 'API_KEY'})
    def usage_plan_id(self, usage_plan_name):
        return self.usage_plans(index_by='name').get(usage_plan_name,{}).get('id')

    def usage_plan_remove_key(self, usage_plan_id, key_id):
         return self._call_method('delete_usage_plan_key', { 'usagePlanId': usage_plan_id,
                                                             'keyId'      : key_id       })

    def usage_plans(self, index_by='id'):
        return self._get('usage_plans', index_by=index_by)

  # 'apiKeyRequired': False,
  # 'authorizationType': 'NONE',
  # 'httpMethod': 'GET',
  # 'methodIntegration': { 'cacheKeyParameters': [],
  #                        'cacheNamespace': '4k04b51pm9',
  #                        'connectionType': 'INTERNET',
  #                        'httpMethod': 'GET',
  #                        'integrationResponses': { '200': { 'responseTemplates': { 'application/json': None},
  #                                                           'statusCode': '200'}},
  #                        'passthroughBehavior': 'WHEN_NO_MATCH',
  #                        'timeoutInMillis': 29000,
  #                        'type': 'HTTP',
  #                        'uri': 'https://www.google.com'},
  # 'methodResponses': { '200': { 'responseModels': {'application/json': 'Empty'},
  #                               'statusCode': '200'}},
  # 'requestParameters': {}}



from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.apis.Lambda                        import Lambda
from osbot_aws.apis.test_helpers.Temp_Aws_Roles   import Temp_Aws_Roles
from osbot_aws.apis.test_helpers.Temp_Folder_With_Lambda_File import Temp_Folder_With_Lambda_File
from osbot_utils.utils.Misc import random_string_and_numbers


class Temp_Lambda:
    def __init__(self, lambda_name=None, lambda_code=None, with_layer=None, delete_on_exit=True):
        self.aws_config        = AWS_Config()
        self.lambda_name       = lambda_name or "temp_lambda_{0}".format(random_string_and_numbers())
        self.aws_lambda        = Lambda(self.lambda_name)
        self.tmp_folder        = Temp_Folder_With_Lambda_File(file_name=self.lambda_name, lambda_code=lambda_code).create_temp_file()
        self.role_arn          = Temp_Aws_Roles().for_lambda_invocation__role_arn() # todo: refactor to have option to create the role programatically (needs feature to wait for role to be available)
        self.create_log        = None
        self.delete_on_exit    = delete_on_exit
        self.wait_max_attempts = 40
        self.s3_bucket         = self.aws_config.lambda_s3_bucket()
        self.s3_key            = 'unit_tests/lambdas/{0}.zip'.format(self.lambda_name)
        self.s3                = self.aws_lambda.s3()
        self.with_layer        = with_layer



    def __enter__(self):
        self.create()
        return self

    def __exit__(self, type, value, traceback):
        if self.delete_on_exit:
            self.delete()

    def arn(self):
        return self.aws_lambda.function_arn()

    def create(self):
        (self.aws_lambda.add_layer      (self.with_layer        )
                        .set_role       (self.role_arn          )
                        .set_s3_bucket  (self.s3_bucket         )
                        .set_s3_key     (self.s3_key            )
                        .set_folder_code(self.tmp_folder.folder )
                        .upload())
        self.create_log = self.aws_lambda.create()
        assert self.create_log.get('status') != 'error'
        update_status   =  self.aws_lambda.wait_for_function_update_to_complete(max_attempts=self.wait_max_attempts)
        assert self.exists() is True
        assert update_status == 'Successful'
        return self.create_log

    def delete(self):
        assert self.aws_lambda.delete()
        self.delete_s3_file()

    def exists(self):
        return self.aws_lambda.exists()

    def info(self):
        return self.aws_lambda.info()

    def invoke(self, params=None):
        return self.aws_lambda.invoke(params)

    def invoke_return_logs(self, params=None):
        return self.aws_lambda.invoke_return_logs(params)

    def invoke_raw(self, params=None):
        return self.aws_lambda.invoke_raw(params)

    def set_s3_bucket(self, value):
        self.s3_bucket = value

    def s3_file_exists(self):
        return self.s3.file_exists(self.s3_bucket, self.s3_key)

    def delete_s3_file(self):
        self.s3.file_delete(self.s3_bucket, self.s3_key)
        assert self.s3_file_exists() is False



from osbot_aws.aws.bedrock.cache.html.Bedrock_Cache__Html import Bedrock_Cache__Html
from osbot_utils.helpers.html.Tag__Base import Tag__Base
from osbot_utils.helpers.html.Tag__H import Tag__H
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json import json_loads, json_dumps
from osbot_utils.utils.Misc import list_set, word_wrap, timestamp_to_str_date


class Bedrock_Cache__Html_Table(Bedrock_Cache__Html):

    def __init__(self):
        super().__init__()
        self.title            = 'AWS Bedrock Cached Data'
        self.sub_title        = 'all entries'
        self.table_headers    = ['model', 'task_type', 'prompt', 'response', 'comments', 'timestamp']
        self.debug_mode       = True
        self.target_file      = '/tmp/tmp-bedrock-table.html'

    def create_table(self, headers, rows):
        tag__table      = Tag__Base(tag_name='table', tag_classes=["table", "table-striped", "table-bordered", "table-hover",'table-dark'])
        tag__thead      = Tag__Base(tag_name='thead')
        tag__tbody      = Tag__Base(tag_name='tbody')

        tag__tr__header = Tag__Base(tag_name='tr')

        for header in headers:
            tag__td = Tag__Base(tag_name='td', inner_html=header)
            tag__tr__header.append(tag__td)

        for row in rows:
            tag__tr = Tag__Base(tag_name='tr')
            for header in headers:
                td_value =  row.get(header)
                tag__td = Tag__Base(tag_name='td', inner_html=td_value)
                tag__tr.append(tag__td)
            tag__tbody.append(tag__tr)

        tag__thead.append(tag__tr__header)
        tag__table.append(tag__thead)
        tag__table.append(tag__tbody)

        self.row_elements.append(tag__table)


    def table_rows(self):
        # if self.debug_mode:
        #     self.table_headers.append('request_body')

        rows    = []
        #pprint(self.table.fields_names__cached())
        for row in self.table.rows():
            comments          = row.get('comments' )
            timestamp         = row.get('timestamp')
            response_data_str = row.get('response_data')
            response_data     = json_loads(response_data_str)
            request_data_str  = row.get('request_data')
            request_data      = json_loads(request_data_str )
            request_body      = request_data.get('body'   , {})
            if not request_body:
                continue
            request_body_str  = json_dumps(request_body)
            model             = request_data.get('model'  ) or request_data.get('model_id'  )       #todo
            task_type         = self.extract_task_type_from_request_body(model, request_body)
            prompt            = self.extract_prompt_from_request_data   (model, request_body)
            response          = self.extract_response_from_response_data(model, response_data)
            if timestamp:
                timestamp_str = timestamp_to_str_date(timestamp)
            else:
                timestamp_str = 'NA'
            if response is None:
                self.extract_response_from_response_data(model, response_data)
                pass
            item = dict(model        = model                                                 ,
                        prompt       = f'<pre>{word_wrap(prompt  .strip(), 40)}</pre>',
                        response     = f'<pre>{word_wrap(response.strip(), 40)}</pre>',
                        task_type    = task_type                                             ,
                        timestamp    = timestamp_str                                         ,
                        comments     = comments                                              )
                        #request_body = f'<pre>{word_wrap(request_body_str, 40)}</pre>')
            # todo: find better way to do this, where we don't see the image base64 string on the html page
            # if self.debug_mode:
            #     item['request_body'] = f'<pre>{word_wrap(request_body_str, 40)}</pre>'
            rows.append(item)
            #pprint(list_set(row))

        return rows

    def extract_task_type_from_request_body(self, model, request_body):
        if 'taskType' in request_body:
            return request_body.get('taskType')
        if model in ['amazon.titan-tg1-large', 'amazon.titan-text-lite-v1']:
            return 'amazon text model'
        if model in ['anthropic.claude-instant-v1','anthropic.claude-v2', 'anthropic.claude-v2:1',
                     "anthropic.claude-3-haiku-20240307-v1:0","anthropic.claude-3-sonnet-20240229-v1:0"]:
            if "prompt" in request_body:
                return "claude text completion"
            if "messages" in request_body:
                return "claude messages"
        #pprint(model, request_body)
        return 'NA'

    def extract_prompt_from_request_data(self, model, request_body):
        if model=='amazon.titan-image-generator-v1':
            task_type = request_body.get('taskType')
            text_params = {}
            if task_type == 'TEXT_IMAGE':
                text_params = request_body.get('textToImageParams'   , {})
            elif task_type == 'IMAGE_VARIATION':
                text_params = request_body.get('imageVariationParams', {})
            if text_params:
                text          = text_params.get('text'            , '')
                negative_text = text_params.get('negativeText'    , '')
                if negative_text:
                    return f'{text}   \n\nnegative_text: {negative_text}'
                return text
            return f'in amazon.titan-image-generator-v1 , unsupported task_type: {task_type}'
        if model in ['amazon.titan-text-lite-v1', 'amazon.titan-tg1-large']:
            return request_body.get('inputText', '')
        if model in ['anthropic.claude-instant-v1','anthropic.claude-v2', 'anthropic.claude-v2:1',
                     "anthropic.claude-3-haiku-20240307-v1:0","anthropic.claude-3-sonnet-20240229-v1:0"]:
            if "prompt" in request_body:
                return request_body.get('prompt', '')
            if "messages" in request_body:
                messages = request_body.get('messages')
                system   = request_body.get('system')
                return f'{messages} \n\nsystem: {system}'
        if model is None:
            return 'NA'
        return f'unsupported model: {model}'

    def extract_response_from_response_data(self, model, response_data):
        if model == 'amazon.titan-image-generator-v1':
            images = list_set(response_data.get('images'))
            if len(images) ==1:
                image = images[0]
                return f'{len(image):,} - image size'
            return f'error: response_data should only have one image and it had {len(images)}'
        if model in ['amazon.titan-text-lite-v1', 'amazon.titan-tg1-large']:
            if type(response_data) is list:
                output_text = ""
                for chunk in response_data:
                    output_text += chunk.get('outputText')
                return output_text
            results = response_data.get('results')
            if len(results) ==1:
                result = results[0]
                output_text = result.get('outputText')
                return output_text
            return f'error: response_data should only have one result and it had {len(results)}'

        if model in ['anthropic.claude-instant-v1','anthropic.claude-v2', 'anthropic.claude-v2:1',
                     "anthropic.claude-3-haiku-20240307-v1:0", "anthropic.claude-3-sonnet-20240229-v1:0"]:
            if "completion" in response_data:
                return response_data.get('completion')
            if "content"  in response_data:
                return f"{response_data.get('content')}"
            if type(response_data) is list:
                completions = ""
                for chunk in response_data:
                    completions += chunk.get('completion')
                return completions

        # if model in ["anthropic.claude-3-haiku-20240307-v1:0", "anthropic.claude-3-sonnet-20240229-v1:0"]:
        #     return f"{response_data.get('content')}"

        return f'unsupported model: {model}'

    def create(self):
        rows    = self.table_rows()
        headers = self.table_headers
        self.create_table(headers, rows)


        #self.append(div_h1)

        self.create_html_file()


from osbot_aws.apis.Lambda import Lambda

LOG_TO_ELK_ENABLED=False

def log_info(message, data = None, index = "gw_bot_logs",category = "API_GS_Bot"):
    return log_to_elk(message=message, data=data, index=index, level='info', category=category)

def log_debug(message, data=None, index="gw_bot_logs", category="API_GS_Bot"):
    return log_to_elk(message=message, data=data, index=index, level='debug', category=category)

def log_error(message, data = None, index = "gw_bot_logs", category = "API_GS_Bot"):
    return log_to_elk(message=message, data=data, index=index, level='error', category=category)

def log_to_elk(message, data = None, index = "gw_bot_logs", level = "debug", category = "API_GS_Bot"):
    payload = {
                "index"    : index    ,
                "level"    : level    ,
                "message"  : message  ,
                "category" : category ,
                "data"     : data
              }
    if LOG_TO_ELK_ENABLED:
        Lambda('gw_bot.lambdas.log_to_elk').invoke_async(payload)
    else:
        print(payload)
    return payload

def slack_message(text, attachments = None, channel = None, team_id=None):  # GBMGMK88Z is the 'from-aws-lambda' channel in the GS-CST Slack workspace
    if attachments is None: attachments = []
    payload = {
                'text'        : text        ,
                'attachments' : attachments ,
                'channel'     : channel     ,
                'team_id'     : team_id
              }
    if channel:
        Lambda('gw_bot.lambdas.slack_message').invoke_async(payload)
    else:
        return text, attachments


def screenshot_from_url(url):
    payload = {"params": ['screenshot', url]}
    return  Lambda('osbot_browser.lambdas.lambda_browser').invoke(payload)




from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.aws.iam.IAM_Assume_Role              import IAM_Assume_Role
from osbot_aws.aws.iam.utils.Temp_Role__For_Service__Config import Temp_Role__For_Service__Config
from osbot_utils.base_classes.Type_Safe             import Type_Safe
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.utils.Env                          import load_dotenv



class Temp_Role__For_Service(Type_Safe):
    aws_config                    : AWS_Config
    _temp_role_config             : Temp_Role__For_Service__Config

    def _iam_assume_role(self):
        role_data         = self._temp_role_config.data()
        service           = role_data.boto3_service_name
        action            = role_data.action
        resource          = role_data.resource
        role_name         = role_data.role_name
        required_services = role_data.required_services

        policies_to_add = []
        for temp_role__service in required_services:
            policies_to_add.append(dict(service=temp_role__service, action=action, resource=resource))

        iam_assume_role = IAM_Assume_Role(role_name=role_name, policies_to_add=policies_to_add)
        return iam_assume_role

    def _create_role_and_credentials(self):
        role_recreate = self._temp_role_config.recreate
        self._iam_assume_role().create_role(recreate=role_recreate)
        return self

    def _config_role_to_create(self, boto3_service_name, required_services):
        self._temp_role_config.boto3_service_name = boto3_service_name
        self._temp_role_config.required_services  = required_services
        return self

    def _temp_role__iam_reset_credentials(self):
        return self._iam_assume_role().credentials_reset()

    # methods that override the normal Boto3 helper classes in OSBot_AWS
    @cache_on_self
    def client(self):
        self._create_role_and_credentials()
        service_name    = self._temp_role_config.boto3_service_name
        iam_assume_role = self._iam_assume_role()
        return iam_assume_role.boto3_client(service_name=service_name)

    @cache_on_self
    def resource(self):
        self._create_role_and_credentials()
        service_name = self._temp_role_config.boto3_service_name
        iam_assume_role = self._iam_assume_role()
        return iam_assume_role.boto3_resource(service_name=service_name)






from os import chmod

import boto3
from botocore.exceptions import ClientError

from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Lists import list_index_by, list_get

from osbot_aws.AWS_Config import set_aws_region, AWS_Config
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Files import file_name, temp_folder, path_combine, file_create

from osbot_utils.decorators.methods.remove_return_value import remove_return_value

from osbot_utils.decorators.lists.group_by import group_by
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.cache import cache
from osbot_aws.apis.Session import Session
from osbot_utils.utils.Misc import list_set
from osbot_utils.utils.Status import status_warning, status_ok

# todo: find good solution to capture/manage these config values
EC2_WAITER_DELAY        = 1             # default was 15 seconds
EC2_WAITER_MAX_ATTEMPTS = 600           # default was 40 times

class EC2(Type_Safe):

    aws_config : AWS_Config

    @cache_on_self
    def client(self):
        return Session().client('ec2')

    @cache_on_self
    def resource(self):                     # todo: refactor this out and only use client() for all calls
        return Session().resource('ec2')

    def ami(self, ami_id):
        #result = self.amis(ami_id=ami_id)
        images = self.client().describe_images(ImageIds=[ami_id]).get('Images')
        if len(images) == 1:
            return images.pop()

    @index_by
    @group_by
    def amis(self, owner='self', architecture=None, state='available', name=None, description=None, ami_id=None, **kwargs):  # todo: find how to search for amis in the quick start
        filters = [{'Name': 'state', 'Values': [state]}]
        if architecture : filters.append({'Name': 'architecture', 'Values': [architecture]})
        if name         : filters.append({'Name': 'name'        , 'Values': [name        ]})
        if description  : filters.append({'Name': 'description' , 'Values': [description ]})
        if ami_id       : filters.append({'Name': 'image-id'    , 'Values': [ami_id      ]})
        for key,value in kwargs.items():
            filters.append({'Name': key, 'Values': [value]})
        describe_kwargs = {'Owners': [owner],
                           'Filters':filters }
        pprint(describe_kwargs)
        return self.client().describe_images(**describe_kwargs).get('Images')

    def image(self, image_id):
        return self.ami(ami_id=image_id)

    def instance_create(self, image_id                                      ,   # only var needed is the image_id which is region specific
                              dry_run               = False                 ,   # todo: refactor into a EC2_Create_Instance helper class
                              block_device_mappings = None                  ,
                              instance_type         = 't2.micro'            ,
                              iam_instance_profile  = None                  ,
                              spot_instance         = False                 ,
                              key_name              = None                  ,
                              name                  = 'created by osbot_aws',
                              network_interface     = None                  ,
                              tags                  = None                  ,
                              security_group_id     = None                  ):


        kwargs = {  "ImageId"      : image_id                                                               ,
                    "InstanceType" : instance_type                                                          ,
                    "MaxCount"     : 1                                                                      ,
                    "MinCount"     : 1                                                                      ,
                    'TagSpecifications': self.tag_specifications_create(tags=tags, name=name,resource_type='instance')}

        if block_device_mappings : kwargs["BlockDeviceMappings"] = block_device_mappings
        if iam_instance_profile  : kwargs["IamInstanceProfile" ] = iam_instance_profile
        if key_name              : kwargs["KeyName"            ] = key_name
        if network_interface     : kwargs['NetworkInterfaces'  ] = [network_interface]
        if security_group_id     : kwargs['SecurityGroupIds'   ] = [security_group_id]
        if spot_instance:
            max_spot_price                  = '0.1'                     # max price to pay
            spot_instance_type              = 'one-time'                # or 'persistent'
            interrupt_behaviour             = 'terminate'               # or 'stop' or 'hibernate'
            kwargs['InstanceMarketOptions'] = { 'MarketType' : 'spot',
                                                'SpotOptions': { 'MaxPrice'                    : max_spot_price      ,
                                                                 'SpotInstanceType'            : spot_instance_type  ,
                                                                 'InstanceInterruptionBehavior': interrupt_behaviour }}
        if dry_run:
            return kwargs

        result   = self.client().run_instances(**kwargs)
        instance = result.get('Instances')[0]
        return  instance.get('InstanceId')

    def create_image(self, instance_id, name, description=None, no_reboot=False, tags=None):
        kwargs = { 'InstanceId' : instance_id ,
                   'Name'       : name        }
        if description: kwargs['Description'] = description
        if no_reboot  : kwargs['NoReboot'   ] = no_reboot
        if tags       : kwargs['TagSpecifications'] = self.tag_specifications_create(tags=tags, resource_type='image')
        return self.client().create_image(**kwargs).get('ImageId')
    def format_instance_details(self, target):
        if target:
            instance_details = { 'architecture'     : target.architecture                         ,
                                 'availability_zone': target.placement.get('AvailabilityZone')    ,
                                 'block_devices'    : {}                                          ,
                                 'cpus'             : target.cpu_options['CoreCount']             ,
                                 'image_id'         : target.image_id                             ,
                                 'instance_type'    : target.instance_type                        ,
                                 'instance_id'      : target.instance_id                          ,
                                 'key_name'         : target.key_name                             ,
                                 'launch_time'      : target.meta.data.get('LauchTime')           ,
                                 'private_ip'       : target.private_ip_address                   ,
                                 'public_ip'        : target.public_ip_address                    ,
                                 'subnet_id'        : target.meta.data.get('SubnetId')            ,
                                 'spot_id'          : target.spot_instance_request_id             ,
                                 'state'            : target.state                                ,
                                 'tags'             : target.tags                                 ,
                                 'vpc_id'           : target.meta.data.get('VpcId')               }

            for block_device_mapping in target.block_device_mappings:
                block_device_name = block_device_mapping.get('DeviceName')
                ebs_data = block_device_mapping.get('Ebs')
                if ebs_data is None:
                    block_device = {"block_type": 'unknown' }               # todo: handle better (including with logging) this scenario (see if there are other possible values for this)
                else:
                    block_device      = {"block_type" : "Ebs"                          ,
                                         "status"     : str(ebs_data.get('Status'    )),
                                         "volume_id"  : str(ebs_data.get('VolumeId'  ))}
                instance_details.get('block_devices')[block_device_name] = block_device

            return instance_details

    def instance_block_devices(self, instance_id):
        instance = self.instance_details_raw(instance_id)
        if instance:
            return list_index_by(instance.block_device_mappings, "DeviceName")

    def instance_details(self, instance_id):
        instance = self.instance_details_raw(instance_id)
        return self.format_instance_details(instance)

    def instance_details_raw(self, instance_id):
        kwargs = { }
        if instance_id: kwargs['InstanceIds'] = [instance_id]
        result = self.resource().instances.filter(**kwargs)             # todo see client().describe_instances is a better method to use
        for instance in result:
            return instance

    def instance_delete(self, instance_id):
        if self.instance_exists(instance_id):
            result = self.client().terminate_instances(InstanceIds=[instance_id])
            return list_get(result.get('TerminatingInstances'),0)

    def instance_exists(self, instance_id):
        return self.instance_details(instance_id=instance_id) is not None

    def instances_details(self, filters=None):
        instances = {}
        if filters is None:
            resource_data = self.resource().instances.all()
        else:
            if type(filters) is not list:
                filters=[filters]
            resource_data = self.resource().instances.filter(Filters=filters)
        for instance in resource_data:
            instance_id = instance.instance_id
            instances[instance_id] = self.format_instance_details(instance)
        return instances

    def instance_start(self, instance_id):
        self.client().start_instances(InstanceIds=[instance_id])

    def instance_stop(self, instance_id):
        self.client().stop_instances(InstanceIds=[instance_id])

    def instance_terminate(self, instance_id):
        self.client().terminate_instances(InstanceIds=[instance_id])

    def instance_volumes_ids(self, instance_id):
        volume_ids = []
        for block_device in self.instance_block_devices(instance_id).values():
            volume_id = block_device.get('Ebs', {}).get('VolumeId')                 # todo: see what other options could exist apart from EBS
            if volume_id:
                volume_ids.append(volume_id)
        return volume_ids

    @index_by
    @group_by
    def instance_volumes(self, instance_id=None):
        volumes_ids = []
        if instance_id:
            volumes_ids = self.instance_volumes_ids(instance_id)
        return self.volumes(volumes_ids)

    def internet_gateway(self, internet_gateway_id):
        result = self.internet_gateways(internet_gateways_ids=[internet_gateway_id])
        if len(result) == 1:
            return result[0]

    def internet_gateway_create(self, tags=None):
        kwargs = { 'TagSpecifications': self.tag_specifications_create(tags=tags, resource_type='internet-gateway') }
        return self.client().create_internet_gateway(**kwargs).get('InternetGateway')

    def internet_gateway_delete(self, internet_gateway_id):
        return self.client().delete_internet_gateway(InternetGatewayId=internet_gateway_id)

    def internet_gateway_exists(self, internet_gateway_id):
        return self.internet_gateway(internet_gateway_id) is not None

    @index_by
    @group_by
    def internet_gateways(self, internet_gateways_ids=None):                         # todo: refactor this entire internet_gateways and (for example) vpns, since they have mostly the same code and there is a number of helper methods added for each of them
        kwargs = {}
        if internet_gateways_ids:
            kwargs['InternetGatewayIds'] = internet_gateways_ids
        try:
            return self.client().describe_internet_gateways(**kwargs).get('InternetGateways')
        except ClientError as e:                                                    # todo: refactor this pattern into a helper with statement (which receives as param the Error.Code value
            if e.response.get('Error', {}).get('Code') == 'InvalidInternetGatewayID.NotFound':
                return []
            raise

    def key_pair(self, key_pair_id=None, key_pair_name=None):
        if key_pair_id:
            return self.key_pairs(index_by="KeyPairId").get(key_pair_id)
        if key_pair_name:
            return self.key_pairs(index_by="KeyName"  ).get(key_pair_name)

    @remove_return_value(field_name='ResponseMetadata')
    def key_pair_create(self, key_name, tags=None):
        kwargs = {'KeyName'          : key_name                                                           ,
                  'TagSpecifications': self.tag_specifications_create(tags=tags, resource_type='key-pair')}

        return self.client().create_key_pair(**kwargs)

    def key_pair_create_to_file(self, key_name, target_folder=None, tags=None):
        key_pair          = self.key_pair_create(key_name=key_name,tags=tags)
        key_pair_id       = key_pair.get('KeyPairId')
        key_pair_material = key_pair.get('KeyMaterial')

        if target_folder is None:
            target_folder = temp_folder()
        path_key_pair = path_combine(target_folder, key_name + ".pem")
        file_create(path_key_pair,key_pair_material)
        chmod(path_key_pair, 0o400)
        return {'path_key_pair':path_key_pair, 'key_pair_id':key_pair_id, 'key_pair':key_pair}

    def key_pair_delete(self, key_pair_id=None, key_pair_name=None):
        if key_pair_id:
            self.client().delete_key_pair(KeyPairId=key_pair_id)
        if key_pair_name:
            self.client().delete_key_pair(KeyName=key_pair_name)
        return self.key_pair_exists(key_pair_id=key_pair_id, key_pair_name=key_pair_name) is False

    def key_pair_exists(self, key_pair_id=None, key_pair_name=None):
        return self.key_pair(key_pair_id=key_pair_id, key_pair_name=key_pair_name) is not None

    @index_by
    @group_by
    def key_pairs(self):
        return self.client().describe_key_pairs().get('KeyPairs')

    def route_create(self, route_table_id, internet_gateway_id, destination_cidr_block='0.0.0.0/0'):
        kwargs = { "DestinationCidrBlock": destination_cidr_block ,
                   "GatewayId"           : internet_gateway_id    ,
                   "RouteTableId"        : route_table_id         }
        return self.client().create_route(**kwargs)

    def route_table(self, route_table_id):
        result = self.route_tables(route_tables_ids=[route_table_id])
        if result and len(result) == 1:
            return result[0]

    @remove_return_value(field_name='ResponseMetadata')
    def route_table_associate(self, route_table_id, subnet_id=None, gateway_id=None):
        kwargs = { 'RouteTableId' : route_table_id}
        if subnet_id:
            kwargs['SubnetId'] = subnet_id
        if gateway_id:
            kwargs['GatewayId'] = gateway_id
        return self.client().associate_route_table(**kwargs)

    def route_table_create(self, vpc_id, tags=None):
        kwargs = { 'TagSpecifications': self.tag_specifications_create(tags=tags, resource_type='route-table'),
                   'VpcId'            : vpc_id }
        result = self.client().create_route_table(**kwargs)
        return result.get('RouteTable')

    def route_table_delete(self, route_table_id):
        return self.client().delete_route_table(RouteTableId=route_table_id)

    def route_table_disassociate(self, association_id):
        return self.client().disassociate_route_table(AssociationId=association_id)

    def route_table_exists(self, route_table_id):
        return self.route_table(route_table_id) is not None

    @index_by
    @group_by
    def route_tables(self, route_tables_ids=None):                         # todo: refactor this entire route_tables and (for example) vpns, since they have mostly the same code and there is a number of helper methods added for each of them
        kwargs = {}
        if route_tables_ids:
            kwargs['RouteTableIds'] = route_tables_ids
        try:
            return self.client().describe_route_tables(**kwargs).get('RouteTables')
        except ClientError as e:                                                    # todo: refactor this pattern into a helper with statement (which receives as param the Error.Code value
            if e.response.get('Error', {}).get('Code') == 'InvalidRouteTableID.NotFound':
                return []
            raise

    def security_group(self, security_group_id=None, security_group_name=None):
        if security_group_id:
            return self.security_groups(index_by="GroupId"  ).get(security_group_id)
        if security_group_name:
            return self.security_groups(index_by="GroupName").get(security_group_name)

    def security_group_authorize_ingress(self, security_group_id, port=None, from_port=None, to_port=None, cidr_ip='0.0.0.0/0', ip_protocol='tcp'):
        if port:
            from_port = port
            to_port   = port

        kwargs = { "CidrIp"     : cidr_ip           ,
                   "IpProtocol" : ip_protocol       ,
                   'GroupId'    : security_group_id ,
                   "FromPort"   : from_port         ,
                   "ToPort"     : to_port           }

        return self.client().authorize_security_group_ingress(**kwargs)


    def security_group_create(self, security_group_name, description, vpc_id=None, tags=None):
        if self.security_group_exists(security_group_name=security_group_name):
            security_group_id =  self.security_group(security_group_name=security_group_name).get('GroupId')            # todo see what is a better way to implement this workflow, since at the moment there are two full calls to the self.security_groups() method
            return status_warning(message=f'Security group already existed: {security_group_name}' , data= { 'security_group_id':security_group_id })
        kwargs = { 'Description'      : description                                                               ,
                   'GroupName'        : security_group_name                                                       ,
                   'TagSpecifications': self.tag_specifications_create(tags=tags, resource_type='security-group') }
        if vpc_id:
            kwargs['VpcId'] = vpc_id

        security_group_id = self.client().create_security_group(**kwargs).get('GroupId')
        return status_ok(message=f'Security group created ok: {security_group_name}', data= { 'security_group_id': security_group_id })

    def security_group_delete(self, security_group_id=None, security_group_name=None):
        if security_group_id:
            self.client().delete_security_group(GroupId=security_group_id)
        if security_group_name:
            self.client().delete_security_group(GroupName=security_group_name)
        return self.security_group_exists(security_group_id=security_group_id, security_group_name=security_group_name) is False

    def security_group_default(self):
        default_vpc_id  = self.vpc_default().get('VpcId')
        default_groups  =  self.security_groups(group_by='GroupName').get('default')
        for default_group in default_groups:
            if default_group.get('VpcId') == default_vpc_id:
                return default_group

    def security_group_exists(self, security_group_id=None, security_group_name=None):
        return self.security_group(security_group_id=security_group_id, security_group_name=security_group_name) is not None

    @index_by
    @group_by
    def security_groups(self):
        return self.client().describe_security_groups().get('SecurityGroups')

    def subnet(self, subnet_id):
        result = self.subnets(subnets_ids=[subnet_id])
        if result and len(result) == 1:
            return result[0]

    def subnet_create(self, vpc_id, cidr_block='172.16.1.0/24', tags=None):
        kwargs = { 'CidrBlock'        : cidr_block                                                       ,
                   'TagSpecifications': self.tag_specifications_create(tags=tags, resource_type='subnet'),
                   'VpcId'            : vpc_id                                                           }
        result = self.client().create_subnet(**kwargs)
        return result.get('Subnet')

    def subnet_delete(self, subnet_id):
        return self.client().delete_subnet(SubnetId=subnet_id)

    @index_by
    def subnet_default_for_az(self, az_index=1):
        return list_get(self.subnets_default_for_az(), az_index)

    def subnet_exists(self, subnet_id):
        return self.subnet(subnet_id) is not None

    @index_by
    @group_by
    def subnets(self, subnets_ids=None):
        kwargs = {}
        if subnets_ids:
            kwargs['SubnetIds'] = subnets_ids
        try:
            return self.client().describe_subnets(**kwargs).get('Subnets')
        except ClientError as e:                                                    # todo: refactor this pattern into a helper with statement (which receives as param the Error.Code value
            if e.response.get('Error', {}).get('Code') == 'InvalidSubnetID.NotFound':
                return []
            raise

    @index_by
    def subnets_default_for_az(self):
        return self.subnets(group_by='DefaultForAz').get(True)

    def tag_specifications_create(self, tags=None, name=None, resource_type=None):
        data =  []
        if name:
            if tags:
                tags['Name'] = name
            else:
                tags = { 'Name': name }
        if tags:
            tag_specification = { 'Tags': [] }
            if resource_type:
                tag_specification['ResourceType'] = resource_type
            for key,value in tags.items():
                tag_specification['Tags'].append({'Key':key, 'Value':value})
            data.append(tag_specification)
        return data

    def volume(self, volume_id=None):
        if volume_id:
            volumes = self.volumes([volume_id])
            if len(volumes) ==1:
                return volumes.pop()

    def volumes(self, volumes_ids=None):
        if volumes_ids:
            return self.client().describe_volumes(VolumeIds=volumes_ids).get('Volumes')
        return []

    def vpc(self, vpc_id):
        result = self.vpcs(vpcs_ids=[vpc_id])
        if len(result) == 1:
            return result[0]

    def vpc_attach_internet_gateway(self, vpc_id, internet_gateway_id):
        return self.client().attach_internet_gateway(VpcId=vpc_id, InternetGatewayId=internet_gateway_id)

    def vpc_detach_internet_gateway(self, vpc_id, internet_gateway_id):
        return self.client().detach_internet_gateway(VpcId=vpc_id, InternetGatewayId=internet_gateway_id)

    def vpc_create(self, cidr_block='172.16.0.0/16', tags=None):
        kwargs = {  'CidrBlock'        : cidr_block                                                    ,
                    'TagSpecifications': self.tag_specifications_create(tags=tags, resource_type='vpc')}
        return self.client().create_vpc(**kwargs).get('Vpc')           # todo: see if it makes sense to add the status_ok message here

    def vpc_delete(self, vpc_id):
        self.client().delete_vpc(VpcId=vpc_id)                          # add rest of normal delete methods

    def vpc_default(self):
        return self.vpcs(index_by='IsDefault').get(True)

    def vpc_exists(self, vpc_id):
        return self.vpc(vpc_id) is not None

    def vpc_attribute_set(self, vpc_id, name, value):
        kvwags = {"VpcId" : vpc_id,
                  name :  {'Value' : value}}
        return self.client().modify_vpc_attribute(**kvwags)


    @index_by
    @group_by
    def vpcs(self, vpcs_ids=None):
        kwargs = {}
        if vpcs_ids:
            kwargs['VpcIds'] = vpcs_ids
        try:
            return self.client().describe_vpcs(**kwargs).get('Vpcs')
        except ClientError as e:                                            # todo: refactor this pattern into a helper with statement (which receives as param the Error.Code value
            if e.response.get('Error', {}).get('Code') == 'InvalidVpcID.NotFound':
                return []
            raise

    def vpcs_ids(self):
        return list_set(self.vpcs(index_by='VpcId'))

    def wait_for(self, waiter_type, kwargs):
        waiter = self.client().get_waiter(waiter_type)
        waiter.config.delay        = EC2_WAITER_DELAY
        waiter.config.max_attempts = EC2_WAITER_MAX_ATTEMPTS
        return waiter.wait(**kwargs)

    def wait_for_instance_exists    (self, instance_id): return self.wait_for('instance_exists'     , {"InstanceIds": [instance_id]})
    def wait_for_instance_status_ok (self, instance_id): return self.wait_for('instance_status_ok'  , {"InstanceIds": [instance_id]})
    def wait_for_instance_running   (self, instance_id): return self.wait_for('instance_running'    , {"InstanceIds": [instance_id]})
    def wait_for_instance_stopped   (self, instance_id): return self.wait_for('instance_stopped'    , {"InstanceIds": [instance_id]})
    def wait_for_instance_terminated(self, instance_id): return self.wait_for('instance_terminated' , {"InstanceIds": [instance_id]})

    def wait_for_vpc_available      (self, vpc_id     ): return self.wait_for('vpc_available', {"VpcIds": [vpc_id]})


def get_EC2_in_region(region_name):                     # this method is not thread save
    set_aws_region(region_name=region_name)         # because this is a global value
    ec2 = EC2()
    ec2.client()                                    # get these value cached (since they depend on the global config value)
    ec2.resource()                                  # get these value cached (since they depend on the global config value)
    return ec2


import calendar
import datetime
import time

from osbot_aws.apis.Boto_Helpers import Boto_Helpers
from osbot_aws.apis.Session import Session
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.utils.Misc import list_set, timestamp_utc_now, wait


class Logs(Boto_Helpers):

    def __init__(self, group_name=None, stream_name=None):  # refactor class to new pattern where these values are not provided here in the ctor
        self.log_group_name = group_name
        self.stream_name    = stream_name
        self._logs          = None

    # helper methods
    def client(self):
        if self._logs is None:
            self._logs = Session().client('logs')
        return self._logs

    # main methods

    def create(self):
        self.group_create()
        self.stream_create()
        return self

    def event_add(self, message, timestamp=None, sequence_token=None):
        if timestamp is None: timestamp = timestamp_utc_now()
        return self.events_add([{'message': message, 'timestamp':timestamp}], sequence_token)

    def events_add(self, events, sequence_token=None):
        try:
            if sequence_token:
                result  = self.client().put_log_events(logGroupName=self.log_group_name,logStreamName=self.stream_name,logEvents=events, sequenceToken=sequence_token)
            else:
                result  = self.client().put_log_events(logGroupName=self.log_group_name,logStreamName=self.stream_name,logEvents=events)
            next_sequence_token = result.get('nextSequenceToken')
            if result.get('rejectedLogEventsInfo') is not None:
                return {'status': 'error', 'nextSequenceToken' : next_sequence_token, 'data': result.get('rejectedLogEventsInfo')}
            return {'status': 'ok', 'nextSequenceToken' : next_sequence_token}
        except Exception as error:
            return {'status': 'error', 'data': '{0}'.format(error)}


    def events(self):
        return self.client().get_log_events(logGroupName=self.log_group_name, logStreamName=self.stream_name)

    def group_create(self):
        if self.group_exists(): return False
        self.client().create_log_group(logGroupName=self.log_group_name)
        return self.group_exists()

    def group_delete(self):
        if self.group_exists() is False: return False
        self.client().delete_log_group(logGroupName=self.log_group_name)
        return self.group_exists() is False

    def group_exists(self):
        return self.group_info() is not None

    def group_events(self, log_stream_prefix=None, start_time=None, end_time=None, filter_pattern=None, limit=None):
        kwargs = {"logGroupName": self.log_group_name}
        if log_stream_prefix:
            kwargs['logStreamNamePrefix'] = log_stream_prefix
        if start_time:
            kwargs['startTime'] = start_time
        if end_time:
            kwargs['endTime'] = end_time
        if filter_pattern:
            kwargs['filterPattern'] = filter_pattern
        if limit:
            kwargs['limit'] = limit

        return self.client().filter_log_events(**kwargs).get('events')

    def group_events_wait_for_pattern(self,log_stream_prefix=None, filter_pattern=None, sleep_for=0.5, max_attempts=20):
        if self.group_wait_until_exists():
            for i in range(0,max_attempts):
                events = self.group_events(log_stream_prefix=log_stream_prefix, filter_pattern=filter_pattern)
                if len(events) > 0:
                    return events
                #print('group_events_wait_for_pattern', i,sleep_for)
                wait(sleep_for)
        return []

    def group_info(self):
        for log_group in self.client().describe_log_groups(logGroupNamePrefix= self.log_group_name).get('logGroups'):
            if log_group.get('logGroupName') == self.log_group_name:
                return log_group
        return None

    def group_messages(self, log_stream_prefix=None, start_time=None, end_time=None, filter_pattern=None, limit=None):
        messages = []
        if self.group_wait_until_exists():
            events = self.group_events(log_stream_prefix=log_stream_prefix,start_time=start_time,end_time=end_time,filter_pattern=filter_pattern, limit=limit)
            for event in events:
                messages.append(event.get('message'))
        return messages

    def group_messages_wait_for_pattern(self, log_stream_prefix=None, filter_pattern=None):
        messages = []
        events = self.group_events_wait_for_pattern(log_stream_prefix=log_stream_prefix,filter_pattern=filter_pattern)
        for event in events:
            messages.append(event.get('message'))
        return messages

    def group_wait_until_exists(self, sleep_for = 0.5, max_attempts = 60):
        for i in range(0,max_attempts):
            if self.group_exists():
                return True
            #print('group_wait_until_exists', i,sleep_for)
            wait(sleep_for)
        return False

    @index_by
    def groups(self):
        return list(self.invoke_using_paginator(self.client(), 'describe_log_groups', 'logGroups'))
        #groups = {}
        #for group in
        #    groups[group.get('logGroupName')] = group
        #return groups

    def groups_names(self):
        return list_set(self.groups(index_by='logGroupName'))

    def name(self):
        return self.log_group_name

    def stream_create(self):
        if self.stream_exists() is True: return False
        self.client().create_log_stream(logGroupName=self.log_group_name, logStreamName= self.stream_name)
        return self.stream_exists()

    def stream_delete(self):
        if self.stream_exists() is False: return False
        self.client().delete_log_stream(logGroupName=self.log_group_name, logStreamName= self.stream_name)
        return self.stream_exists() is False

    def stream_exists(self):
        return len(self.streams()) > 0

    def streams(self):
        if self.stream_name is None:
            return self.client().describe_log_streams(logGroupName=self.log_group_name).get('logStreams')
        return self.client().describe_log_streams(logGroupName=self.log_group_name, logStreamNamePrefix=self.stream_name).get('logStreams')

    def messages(self,limit=10000, hours=24, start_time=None, end_time=None, max_loop=25):

        if start_time is None:
            start_time = int((datetime.datetime.now() - datetime.timedelta(hours=hours)).timestamp() * 1000)

        if end_time is None:
            end_time = int(time.time() * 1000)

        # kwargs = dict(  logGroupName  = self.log_group_name ,
        #                 logStreamName = self.stream_name    ,
        #                 limit         = limit               ,
        #                 startTime     = start_time          ,
        #                 endTime       = end_time            )
        # response  = self.client().get_log_events(**kwargs)
        # #events = self.invoke_using_paginator(self.client(), 'get_log_events', 'events', **kwargs)


        params = {
            "logGroupName": self.log_group_name ,
            "logStreamName": self.stream_name   ,
            "startFromHead": True               ,
            "startTime"    : start_time         ,
            "endTime"      : end_time           ,
            "limit"        : limit              }

        log_events = []
        import logging
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        while max_loop > 0:
            logger.info(f"Fetching log events with params: {params}")
            response = self.client().get_log_events(**params)
            events   = response["events"]
            log_events.extend(events)
            next_token = response.get("nextForwardToken")
            logger.info(f"Got {len(events)} events with next_token: {next_token}")
            #print(f'limit:{limit} len(log_events):{len(log_events)}')
            if len(events) == 0:
                break
            if next_token == params.get("nextToken"):
                break
            if len (log_events) >= limit:
                break
            params['nextToken'] = next_token
            max_loop -=1

        messages = []
        for event in log_events:
            messages.append(event.get('message'))
        return messages

    def __repr__(self):
        return f"Logs: {self.log_group_name or ''} {self.stream_name or ''}"


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

class Cohere_Command(Kwargs_To_Self):
    model_id           : str
    prompt             : str
    temperature        : float    = 0.0
    p                  : float    = 1.0
    k                  : int      = 0
    max_tokens         : int      = 1024
    stop_sequences     : list
    return_likelihoods : str      = 'NONE' #"GENERATION|ALL|NONE",
    stream             : bool     = False
    num_generations    : int      = 1
    truncate           : str      = "NONE"      # NONE|START|END"
    #logit_bias        : dict       #    {token_id: bias},



    def body(self):
        return dict(prompt             = self.prompt             ,
                    temperature        = self.temperature        ,
                    p                  = self.p                  ,
                    k                  = self.k                  ,
                    max_tokens         = self.max_tokens         ,
                    stop_sequences     = self.stop_sequences     ,
                    return_likelihoods = self.return_likelihoods ,
                    stream             = self.stream             ,
                    num_generations    = self.num_generations    ,
                    truncate           = self.truncate           )



from osbot_aws.aws.bedrock.models.titan.base_classes.Amazon_Titan__Text import Amazon_Titan__Text

# for details on this model see https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-titan-text.html

class Amazon_Titan_Text_Lite_V1(Amazon_Titan__Text):
    # config vars
    model_id     : str = 'amazon.titan-text-lite-v1'



from osbot_aws.aws.dynamo_db.Dynamo_DB import Dynamo_DB
from osbot_aws.aws.dynamo_db.Dynamo_DB__Table import Dynamo_DB__Table
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self


class Dynamo_DB__Streams(Kwargs_To_Self):

    dynamo_db_table : Dynamo_DB__Table






import zlib

from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Json import json_dumps
from osbot_utils.utils.Misc import str_to_bytes


class Dynamo_DB__Record__Metadata(Kwargs_To_Self):
    created_by       : str
    timestamp_created: int
    timestamp_updated: int
    updated_by       : str

class Dynamo_DB__Record(Kwargs_To_Self):

    data       : dict
    data_binary: bytes
    key_value  : str
    metadata   : Dynamo_DB__Record__Metadata

    def contents(self):
        return self.__locals__()

    def compress_binary_data(self):
        """Compresses the binary data stored in data_binary using zlib with maximum compression."""
        if self.data_binary is not None:
            self.data_binary = zlib.compress(self.data_binary, level=9)  # Level 9 for maximum compression
        return self

    def uncompress_binary_data(self):
        """Decompresses the binary data stored in data_binary using zlib."""
        if self.data_binary is not None:
            self.data_binary = zlib.decompress(self.data_binary)
        return self

    def set_binary_data__from_dict(self, data_as_str):
        data_as_str = json_dumps(data_as_str, pretty=False)
        self.data_binary = str_to_bytes(data_as_str)
        return self

    def set_binary_data__from_str(self, data_as_str):
        self.data_binary = str_to_bytes(data_as_str)
        return self




from osbot_utils.utils.Process import Process

class Shell_Server:

    def invoke(self, data):
        method_name   = data.get('method_name')
        method_kwargs = data.get('method_kwargs')
        if method_name:
            if hasattr(Shell_Server,method_name):
                method = getattr(Shell_Server, method_name)
                if type(method_kwargs) is dict:
                    return_value = method(**method_kwargs)
                else:
                    return_value = method()
                return { "method_invoked" : True          ,
                         "method_name"    : method_name   ,
                         "method_kwargs"  : method_kwargs ,
                         "return_value"   : return_value  }
        return None


    @staticmethod  # note: this method by design allows extra commands injection via | and ;
    def bash(command, cwd=None):
        bash_command = 'bash'
        bash_params  = ['-c', command]
        return Shell_Server.process_run(bash_command, bash_params, cwd)


    @staticmethod
    def process_run(executable, params=None, cwd='.'):
        return Process.run(executable,params, cwd)

    @staticmethod
    def ping():
        return 'pong'

    @staticmethod
    def python_exec(code):
        try:
            exec(code)
            return locals().get('result')
        except Exception as error:
            return {'error': f'{error}'}

    # helper process_run methods
    @staticmethod
    def pwd():
        return Shell_Server.process_run('pwd').get('stdout')

    @staticmethod
    def disk_space():
        return Shell_Server.bash('df -h').get('stdout').strip()

    @staticmethod
    def file_contents(path):
        return Shell_Server.process_run('cat',[path]).get('stdout')

    @staticmethod
    def list_processes():
        return Shell_Server.bash('ps -aux').get('stdout').strip()

    @staticmethod
    def memory_usage():
        return Shell_Server.file_contents('/proc/meminfo')
        #as.asd()
        #return Shell_Server.bash('cat /proc/meminfo').get('stdout')



from osbot_aws.aws.s3.S3                import S3
from osbot_aws.aws.s3.S3__Minio         import S3__Minio
from osbot_utils.testing.Temp_Env_Vars  import Temp_Env_Vars
from unittest                           import TestCase
from osbot_utils.testing.Hook_Method    import Hook_Method
from osbot_utils.utils.Env              import get_env

TEST__AWS_ACCOUNT_ID     = '111122223333'
TEST__AWS_DEFAULT_REGION = 'eu-west-1'


class TestCase__Minio(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.aws_account_id     = get_env('AWS_ACCOUNT_ID'    , TEST__AWS_ACCOUNT_ID    )        # use existing values if they already exist
        cls.aws_default_region = get_env('AWS_DEFAULT_REGION', TEST__AWS_DEFAULT_REGION)
        tmp_vars_values        = dict(AWS_ACCOUNT_ID     = cls.aws_account_id    ,
                                      AWS_DEFAULT_REGION = cls.aws_default_region)
        cls.tmp_env_vars       = Temp_Env_Vars(env_vars=tmp_vars_values).set_vars()
        cls.hook_method        = Hook_Method(S3, 'client'      )       # todo: see if we also need to hook the main S3 class (from osbot_aws)
        cls.hook_method.mock_call = cls.s3_minio__s3_client
        cls.hook_method.wrap()

    @classmethod
    def tearDownClass(cls):
        cls.hook_method.unwrap()
        cls.tmp_env_vars.restore_vars()

    def s3_minio__s3_client(self):
        return S3__Minio().s3_client()


from osbot_aws.aws.bedrock.models.diffusion.Stability_Stable_Diffusion import Stability_Stable_Diffusion
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Misc import list_set
from osbot_utils.utils.Png import save_png_base64_to_file


class Stability_Stable_Diffusion_XL_V1(Stability_Stable_Diffusion):
    model_id            : str      = 'stability.stable-diffusion-xl-v1'
    samples             : int      = 1
    #clip_guidance_preset: str      = 'NONE'         # FAST_BLUE, FAST_GREEN, NONE, SIMPLE SLOW, SLOWER, SLOWEST.
    #sampler             : str                       # DDIM, DDPM, K_DPMPP_2M, K_DPMPP_2S_ANCESTRAL, K_DPM_2, K_DPM_2_ANCESTRAL, K_EULER, K_EULER_ANCESTRAL, K_HEUN K_LMS.
    #style_preset        : str                       # 3d-model, analog-film, anime, cinematic, comic-book, digital-art, enhance, fantasy-art, isometric, line-art, low-poly, modeling-compound, neon-punk, origami, photographic, pixel-art, tile-texture
    #extras              : dict

    def body(self):
        return dict(text_prompts            = self.text_prompts        ,
                    cfg_scale               = self.cfg_scale           ,
                    steps                   = self.steps               ,
                    seed                    = self.seed                ,
                    samples                 = self.samples             ,
                    #clip_guidance_preset    = self.clip_guidance_preset,
                    # sampler               = self.# sampler           ,
                    #style_preset            = self.style_preset        ,
                    #extras                  = self.extras
                    )

    def invoke(self, bedrock, prompts):
        self.set_text_prompts(prompts)
        response  = bedrock.model_invoke(self.model_id, self.body())
        if response:
            artifacts = response.get('artifacts')
            if list_set(response) == ['artifacts', 'result']:
                if len(artifacts) == 1:
                    artifact = artifacts[0]
                    return artifact.get('base64')

    def comments(self, cache):
        model_id = self.model_id
        body = self.body()
        return cache.cache_entry_comments(model_id=model_id, body=body)





from osbot_aws.apis.Cloud_Watch_Logs import Cloud_Watch_Logs
from osbot_aws.aws.ecs.ECS           import ECS


class ECS_Actions:

    def __init__(self):
        self.ecs          = ECS()
        self.logs         = Cloud_Watch_Logs()
        self.account_id   = self.ecs.account_id
        self.region       = self.ecs.region
        self.cluster_name = 'test_ecs_cluster_2'
        self.task_family  = 'test_task_create'
        self.cluster_arn  = self.ecs.cluster_arn(self.cluster_name)

    def create_cluster(self):
        return self.ecs.cluster_create(self.cluster_name)

    def delete_cluster(self):
        return self.ecs.cluster_delete(self.cluster_arn)

    def info_cluster(self):
        return self.ecs.clusters()



from osbot_aws.utils.Version import Version, version
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Misc import date_time_now


class WS__Connection(Kwargs_To_Self):
    # auto populated
    id                  : str  # to be set by Dynamo_DB__Table
    timestamp           : int  # to be set by the request
    version             : str  # set in __init__

    # indexes
    date                : str          # also populated in __init__
    status              : str = 'NA'

    # other fields
    env                 : str
    events              : list
    timestamp_disconnect: int

    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.date    = self.date_today()   # make sure date field is set
        self.version = version             # make all requests capture the code version

    def date_today(self):
        return date_time_now(date_time_format='%Y-%m-%d')


from osbot_utils.utils.Files     import Files, file_write, file_exists
from osbot_utils.utils.Functions import function_source_code
from osbot_utils.utils.Zip import folder_zip


class Temp_Folder_With_Lambda_File:
    def __init__(self, file_name=None, lambda_code=None):
        self.file_name    = file_name or 'simple_lambda.py'
        self.folder       = None
        self.lambda_code  = lambda_code or "def run(event, context): print('OSBot inside an Lambda :)'); return 'hello {0}'.format(event.get('name'))"
        self.tmp_file     = None
        self.create_temp_file()

    def __enter__(self):
        return self.create_temp_file()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete()

    def create_temp_file(self, new_code=None):
        self.folder      = Files.temp_folder('tmp_lambda_')
        self.lambda_code = new_code or function_source_code(self.lambda_code)                   # extracts source code if self.lambda_code is a function
        self.tmp_file    = Files.path_combine(self.folder, '{0}.py'.format(self.file_name))
        file_write(self.tmp_file, self.lambda_code)
        assert file_exists(self.tmp_file)
        return self

    def delete(self):
        return Files.folder_delete_all(self.folder)

    def zip(self):
        return folder_zip(self.folder)

    def zip_bytes(self):
        with open(self.zip(), 'rb') as file_data:
            return file_data.read()


from osbot_aws.aws.dynamo_db.domains.DyDB__Table import DyDB__Table
from osbot_utils.decorators.lists.group_by import group_by
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.utils.Misc import list_set


class DyDB__Table_With_GSI(DyDB__Table):

    def attribute_definitions(self):
        return self.info().get('AttributeDefinitions', [])

    def can_update_table(self):
        if self.table_status() == 'ACTIVE':
            if self.indexes_status() in [['ACTIVE'], []]:
                return True
        return False

    def index(self, index_name):
        return self.indexes__by_index_name().get(index_name)

    def index_create(self, index_name, index_type, sort_key='timestamp', sort_key_type='N', sort_key_schema='RANGE', projection_type='ALL'):
        kwargs = dict(partition_name  = index_name      ,
                      partition_type  = index_type      ,
                      sort_key        = sort_key        ,
                      sort_key_type   = sort_key_type   ,
                      sort_key_schema = sort_key_schema ,
                      projection_type = projection_type )
        result = self.gsi_create(**kwargs).get('data')              # todo: add check for status =='ok'
        return result                                               # todo improve the data returned by this method
        # table_description = result.get('TableDescription'      , {})
        # table_status      = table_description.get('TableStatus')
        # return table_status

    def index_delete(self, index_name):
        return self.gsi_delete(index_name).get('data')

    def index_exists(self, index_name):
        return index_name in self.indexes_names()

    def index_not_exists(self, index_name):
        return self.index_exists(index_name) is False

    @group_by
    @index_by
    def indexes(self):
        return self.info().get('GlobalSecondaryIndexes', [])

    def indexes__by_index_name(self):
        return self.indexes(index_by='IndexName')

    def indexes_names(self):
        return list_set(self.indexes__by_index_name())

    def indexes_status(self):
        return list_set(self.indexes(index_by='IndexStatus'))

    def table_status(self):
        return self.info().get('TableStatus')


from osbot_utils.decorators.methods.cache import cache
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.Session import Session

class Kinesis:

    @cache_on_self
    def kinesis(self):
        return Session().client('kinesis')

    def streams(self):
        return self.kinesis().list_streams()



from osbot_aws.aws.ec2.EC2_Instance import EC2_Instance


class Temp_EC2_Instance:

    def __init__(self, create_kwargs=None, delete_on_exit=True):
        self.ec2_instance   = None
        self.create_kwargs  = create_kwargs
        self.delete_on_exit = delete_on_exit

    def __enter__(self):
        self.ec2_instance = EC2_Instance(create_kwargs=self.create_kwargs)
        self.ec2_instance.create()
        return self.ec2_instance

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.delete_on_exit:
            self.ec2_instance.delete()
        if exc_type is not None:
            raise exc_val



from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self


class IAM__Statement__Resource(Kwargs_To_Self):
    actions  : list
    effect   : str
    resource : str

    def validate_data(self):
        return self.actions and self.effect and self.resource

    def statement(self):
        if self.validate_data():
            action = [action for action in self.actions]
            return {"Effect"  : self.effect,
                    "Action"  : action,
                    "Resource": self.resource}


from osbot_aws.helpers.VPC import VPC
from osbot_utils.utils.Misc import random_string

from osbot_aws.apis.EC2 import EC2


class Temp_VPC:
    def __init__(self, add_internet_gateway=False, add_route_table=False, add_security_group=False, add_subnet=False):
        self.add_internet_gateway = add_internet_gateway
        self.add_route_table      = add_route_table
        self.add_security_group   = add_security_group
        self.add_subnet           = add_subnet
        self.vpc_name             = random_string(prefix='osbot_aws-temp_vpc-')
        self.tags                 = {'Name': self.vpc_name}
        self.vpc                  = VPC(vpc_name=self.vpc_name, tags=self.tags)


    def __enter__(self):
        self.create_vpc()
        self.create_route_table()
        self.create_internet_gateway()
        self.create_security_group()
        self.create_subnet()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.delete_internet_gateway()
        self.delete_route_table()
        self.delete_security_group()
        self.delete_subnet()
        self.delete_vpc()

    def create_internet_gateway(self):
        if self.add_internet_gateway:
            self.vpc.create_internet_gateway()
        return self

    def create_route_table(self):
        if self.add_route_table:
            self.vpc.create_route_table()
        return self

    def create_security_group(self):
        if self.add_security_group:
            self.vpc.create_security_group()
        return self

    def create_subnet(self):
        if self.add_subnet:
            self.vpc.create_subnet()
        return self

    def create_vpc(self):
        self.vpc.create_vpc()
        return self

    def delete_internet_gateway(self):
        if self.add_internet_gateway:
            self.vpc.delete_internet_gateway()
        return self

    def delete_security_group(self):
        if self.add_security_group:
            self.vpc.delete_security_group()
        return self

    def delete_subnet(self):
        if self.add_subnet:
            self.vpc.delete_subnet()
        return self

    def delete_route_table(self):
        if self.add_route_table:
            self.vpc.delete_route_table()
        return self

    def delete_vpc(self):
        self.vpc.delete_vpc()
        return self



from functools import cache

from botocore.exceptions import ClientError

from osbot_aws.apis.Session import Session
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.lists.group_by import group_by
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json import json_dumps, json_loads


class Bedrock(Kwargs_To_Self):
    region_name : str

    def __enter__(self                           ): return self
    def __exit__ (self, exc_type, exc_val, exc_tb): pass

    @cache_on_self
    def client(self):
        session = Session(region_name=self.region_name)
        return session.client('bedrock')

    def model(self, model_id):
        return self.models__by_id().get(model_id)

    def model_invoke(self, model_id, body):
        body_as_str = json_dumps(body)
        accept      = 'application/json'
        contentType = 'application/json'

        response = self.runtime().invoke_model(body=body_as_str, modelId=model_id, accept=accept, contentType=contentType)
        response_body = json_loads(response.get('body').read())
        return response_body

    def model_invoke_stream(self, model_id, body):
        body_as_str = json_dumps(body)
        accept      = 'application/json'
        contentType = 'application/json'

        response = self.runtime().invoke_model_with_response_stream(body=body_as_str, modelId=model_id, accept=accept, contentType=contentType)

        stream   = response.get('body')
        if stream:
            for event in stream:
                chunk = event.get('chunk')
                if chunk:
                    data = json_loads(chunk.get('bytes').decode())
                    yield data

    @index_by
    @group_by
    def models(self):
        response = self.client().list_foundation_models()
        models   = response.get("modelSummaries")
        return models

    @cache_on_self
    def models_active(self):
        models = {}
        for model in self.models():
            model_id   = model.get("modelId"                )
            provider   = model.get("providerName"           )
            for output in model.get('outputModalities'       ):
                for throughput in model.get('inferenceTypesSupported'):
                    status   = model.get('modelLifecycle').get('status')
                    if status == 'ACTIVE':
                        throughputs = models      [throughput] = models     .get(throughput, {})
                        providers   = throughputs [provider  ] = throughputs.get(provider  , {})
                        outputs     = providers   [output    ] = providers  .get(output    , {})
                        outputs[model_id] = model
        return models

    @cache_on_self
    def models__by_id(self):
        return self.models(index_by='modelId')

    def models_by_throughput(self, throughput_mode):
        models_active = self.models_active()
        models        = {}
        tree  = models_active.get(throughput_mode, {})
        for provider in tree.values():
            for _, outputs in provider.items():
                for model_id, model in outputs.items():
                    models[model_id] = model
        return models

    def models_by_throughput__on_demand(self):
        return self.models_by_throughput('ON_DEMAND')

    def models_by_throughput__provisioned(self):
        return self.models_by_throughput('PROVISIONED')

    @cache_on_self
    def runtime(self):
        session = Session(region_name=self.region_name)
        return session.client('bedrock-runtime')


from osbot_aws.apis.Cloud_Watch_Logs        import Cloud_Watch_Logs
from osbot_utils.decorators.lists.index_by  import index_by
from osbot_aws.AWS_Config                   import AWS_Config
from osbot_aws.aws.ec2.EC2                  import EC2
from osbot_aws.aws.ecs.ECS                  import ECS
from osbot_aws.aws.iam.IAM                  import IAM


class ECS_Fargate_Task:

    def __init__(self, cluster_name, image_name, subnet_id=None, security_group_id=None):
        self.ecs                   = ECS()
        self.ec2                   = EC2()
        self.cloud_watch_logs      = Cloud_Watch_Logs()
        self.aws_config            = AWS_Config()                                         # load config from env variables
        self.account_id            = self.cloud_watch_logs.account_id
        self.region_name           = self.cloud_watch_logs.region_name
        self.cluster_name          = cluster_name
        self.image_name            = image_name
        self.subnet_id             = subnet_id
        self.security_group_id     = security_group_id
        self.task_family           = f"family__{self.image_name}"
        self.task_name             = f'task__{self.cluster_name}'
        self.launch_type           = 'FARGATE'
        #self.iam_execution_role = f'fargate-execution-role_{self.region_name}_{self.task_family}'
        #self.iam_task_role      = f'fargate-task-role_{self.region_name}_{self.task_family}'
        self.task_arn              = None
        self.constraint_expression = None

    # helper methods
    def cluster_arn(self):
        return self.ecs.cluster_arn(cluster_name=self.cluster_name)

    def delete_cluster_task_definition_tasks_roles(self):
        # stop task
        self.stop()
        self.wait_for_task_stopped()

        # delete task_definitions
        task_definitions_arns = self.ecs.task_definitions_arns(task_family=self.task_family)
        for task_definition_arn in task_definitions_arns:
            self.ecs.task_definition_delete(task_definition_arn=task_definition_arn)


        # delete roles
        task_definition_config = self.ecs.task_definition_setup(task_family=self.task_family, image_name=self.image_name)
        execution_role_name    = task_definition_config.get('execution_role_name')
        task_role_name         = task_definition_config.get('task_role_name')
        IAM(role_name=execution_role_name).role_delete()
        IAM(role_name=task_role_name     ).role_delete()

        # delete cluster
        self.ecs.cluster_delete(cluster_arn=self.cluster_arn())

    def setup(self):
        self.setup_cluster()
        self.setup_task_definition()
        return self

    def setup_cluster(self):
        self.ecs.cluster_create                  (cluster_name = self.cluster_name      )
        #self.ecs.policy_create_for_task_role     (role_name    = self.iam_task_role     )
        #self.ecs.policy_create_for_execution_role(role_name    = self.iam_execution_role)

    def setup_task_definition(self):
        task_definition_arn = self.task_definition_arn()

        if self.ecs.task_definition_exists(task_definition_arn=task_definition_arn):
            return {}                                                                   # todo: find better return value and pattern for this workflow
        task_definition_config = self.ecs.task_definition_setup(task_family=self.task_family,image_name=self.image_name)
        task_definition        = self.ecs.task_definition_create(task_definition_config=task_definition_config)
        return { "task_definition_config": task_definition_config, "task_definition": task_definition }

    def stop(self):
        return self.ecs.task_stop(cluster_name=self.cluster_name, task_arn=self.task_arn)

    def task_config(self):
        task_config = { "cluster_name"          : self.cluster_name                                                          ,
                        "security_group_id"     : self.security_group_id or self.ec2.security_group_default().get('GroupId' ),
                        "subnet_id"             : self.subnet_id         or self.ec2.subnet_default_for_az ().get('SubnetId'),
                        "task_definition_arn"   : self.task_definition_arn()                                                 ,
                        "launch_type"           : self.launch_type                                                           ,
                        'constraint_expression' : self.constraint_expression                                                 }

        return task_config

    def task_definition(self):
        return self.ecs.task_definition(task_family=self.task_family)

    def task_definition_arn(self):
        return self.task_definition().get('taskDefinitionArn')

    # main methods

    def create(self, task_config=None):
        if task_config is None:
            task_config = self.task_config()

        task          = self.ecs.task_create(**task_config)
        self.task_arn = task.get('taskArn')

        return task                                                                     # return status_ok and capture potential error

    def container(self):
        return self.ecs.container(task_arn=self.task_arn , image_name=self.image_name, cluster_name=self.cluster_name)

    def container_definition(self):
       return self.ecs.container_definition(task_definition_arn=self.task_definition_arn, image_name=self.image_name)

    @index_by
    def containers(self):
        return self.ecs.containers(task_arn=self.task_arn ,cluster_name=self.cluster_name)

    @index_by
    def containers_definitions(self):
        return self.ecs.containers_definitions(task_definition_arn=self.task_definition_arn)

    def info(self):
        return self.ecs.task(cluster_name=self.cluster_name, task_arn=self.task_arn)

    def log_stream_config(self):
        return self.ecs.log_stream_config(task_arn=self.task_arn, task_definition_arn=self.task_definition_arn(), image_name=self.image_name, cluster_name=self.cluster_name)

    def log_stream_exists(self):
        return self.ecs.log_stream_exists(task_arn=self.task_arn, task_definition_arn=self.task_definition_arn(), image_name=self.image_name, cluster_name=self.cluster_name)

    def logs(self):
        return self.ecs.logs(task_arn=self.task_arn, task_definition_arn=self.task_definition_arn(), image_name=self.image_name, cluster_name=self.cluster_name)

    def logs_wait_for_data(self, max_attempts=60 ,sleep_for=1):
        return self.ecs.logs_wait_for_data(task_arn=self.task_arn, task_definition_arn=self.task_definition_arn(), image_name=self.image_name, cluster_name=self.cluster_name,max_attempts=max_attempts ,sleep_for=sleep_for)

    def status(self):
        return self.info().get('lastStatus')

    def wait_for_task_running(self):
        return self.ecs.wait_for_task_running(cluster_name=self.cluster_name, task_arn=self.task_arn)

    def wait_for_task_stopped(self):
        return self.ecs.wait_for_task_stopped(cluster_name=self.cluster_name, task_arn=self.task_arn)


import random
import string
from osbot_utils.testing.Temp_Env_Vars import Temp_Env_Vars

class Temp__Random__AWS_Credentials(Temp_Env_Vars):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        empty_aws_credentials = dict(                           # Create the AWS credentials with realistic random values
            AWS_ACCESS_KEY_ID       =   self.random_aws_access_key_id    (),
            AWS_SECRET_ACCESS_KEY   =   self.random_aws_secret_access_key(),
            AWS_ACCOUNT_ID          =   self.random_aws_account_id       (),
            AWS_DEFAULT_REGION      =   self.random_aws_region           ()
        )
        self.env_vars.update(empty_aws_credentials)                                 # add temp aws vars to exiting env_vars (set in ctor)


    # Helper functions to generate random values matching AWS schema

    def random_aws_access_key_id(self):
        return 'AKIA' + ''.join(random.choices(string.ascii_uppercase + string.digits, k=16))       # AWS access key IDs are typically 20 characters long, uppercase letters and digits

    def random_aws_secret_access_key(self):
        return ''.join(random.choices(string.ascii_letters + string.digits + '/+=', k=40))          # AWS secret access keys are typically 40 characters long, mix of letters, digits, and special characters

    def random_aws_account_id(self):
        return ''.join(random.choices(string.digits, k=12))                                         # AWS account IDs are typically 12 digits long

    def random_aws_region(self):
        regions = [ 'us-east-1'     , 'us-west-1'   , 'us-west-2'     ,
                    'eu-west-1'     , 'eu-central-1', 'ap-southeast-1',
                    'ap-northeast-1', 'ap-south-1'  , 'sa-east-1'     ]
        return random.choice(regions)                                                               # Randomly select from a list of common AWS regions



from osbot_utils.utils.Misc         import random_string
from osbot_utils.utils.Status import status_error, status_ok

from osbot_aws.AWS_Config           import AWS_Config
from osbot_aws.aws.iam.IAM          import IAM
from osbot_aws.aws.iam.IAM_Policy   import IAM_Policy


class IAM_Role:
    def __init__(self,role_name=None, role_arn=None):
        if role_arn:
            self.role_name = role_arn.split('/')[-1]
        else:
            self.role_name  = role_name or f"osbot_temp_role_{random_string()}"
        self.iam        = IAM(role_name=self.role_name)

    def add_policy_for__lambda(self):
        temp_policy_name = 'policy_{0}'.format(self.role_name)
        cloud_watch_arn  = f'arn:aws:logs:{AWS_Config().aws_session_region_name()}:{AWS_Config().aws_session_account_id()}:log-group:/aws/lambda/*'
        iam_policy       = IAM_Policy(temp_policy_name)
        policy_arn       = iam_policy.add_cloud_watch(cloud_watch_arn).create().get('policy_arn')
        self.iam.role_policy_attach(policy_arn)
        return policy_arn

    def arn(self):
        return self.iam.role_arn()

    def attach_policy(self, policy_name, policy_document):
        self.delete_policy(policy_name= policy_name)
        result_create = self.iam.policy_create(policy_name=policy_name, policy_document=policy_document)
        policy_arn = result_create.get('policy_arn')
        self.iam.role_policy_attach(policy_arn=policy_arn)
        return policy_arn

    def create(self, assume_policy_document, skip_if_exists=True):
        try:
            role_info = self.iam.role_create(assume_policy_document=assume_policy_document, skip_if_exists=skip_if_exists)
            return status_ok(data    = role_info                                         ,
                             message = f'user created (skip_if_exists: {skip_if_exists})')
        except Exception as error:
            return status_error(data    = dict(assume_policy_document=assume_policy_document)                         ,
                                error   = error                                                                       ,
                                message = f'error creating role with assume_policy_document: {assume_policy_document}')

        #return self.exists()

    def create_for__lambda(self):
        result = self.create_for_service__assume_role('lambda.amazonaws.com')
        if result.get('status') == 'ok':
            self.add_policy_for__lambda()
        return result

    def create_for__code_build(self):
        return self.create_for_service__assume_role('codebuild.amazonaws.com')

    def create_for_service__assume_role(self, service):
        statement = {'Action': 'sts:AssumeRole',
                      'Effect': 'Allow',
                      'Principal': {'Service': service}}
        return self.create_from_statement(statement)

    def create_for_service(self, service, statement):
        statement['Principal'] = {'Service': service}
        return self.create_from_statement(statement)

    def create_for_service_with_policies(self, service, policies, project_name, recreate_policy = False):
        role          = self.create_for_service__assume_role(service)
        role_arn      = role.get('role_arn')
        policies_arns = self.iam.policies_create(policies, project_name, recreate_policy)
        self.iam.role_policies_attach(policies_arns)
        return { "role_arn":  role_arn, "policies_arns" :policies_arns }

    def create_from_statement(self, statement):
        return self.create_from_statements([statement])

    def create_from_statements(self, statement):
        role_arn =  self.iam.role_arn()
        if role_arn:
            return {'status':'warning', 'data': 'role already exists', 'role_name': self.iam.role_name , 'role_arn': role_arn}
        else:
            policy_document = {'Statement': statement}
            data = self.iam.role_create(policy_document)
            return {'status': 'ok', 'data': data, 'role_name': self.iam.role_name, 'role_arn': data.get('Arn') }

    def create_instance_profile(self):
        return self.iam.role_create_instance_profile(role_name=self.role_name).get('InstanceProfile')

    def add_to_instance_profile(self):
        return self.iam.role_add_to_instance_profile(instance_profile_name=self.role_name, role_name=self.role_name) # it seems that the convension is to use the same for both

    def delete(self):
        return self.iam.role_delete()

    def delete_policy(self, policy_arn=None, policy_name=None):
        return self.iam.policy_delete(policy_arn=policy_arn, policy_name=policy_name)

    def exists(self):
        return self.iam.role_exists()

    def info(self):
        return self.iam.role_info()

    def not_exists(self):
        return self.iam.role_not_exists()

    def policies(self):
        for policy_name, policy_statements in self.policies_statements().items():
            yield IAM_Policy(policy_name, policy_statements)


    def policies_statements(self):
        return self.iam.role_policies_statements()


import requests

from osbot_aws.apis.shell.Shell_Client import Shell_Client

DOCKER__LAMBDA__INVOKE_URL ="http://localhost:9000/2015-03-31/functions/function/invocations"

class Shell_Client__Lambda__Local(Shell_Client):

    def __init__(self, invoke_url=None):
        self.invoke_url = invoke_url or DOCKER__LAMBDA__INVOKE_URL

    def _invoke(self, method_name, method_kwargs=None, return_logs=False):
        auth_key = self._lambda_auth_key()
        payload = {'lambda_shell': {'method_name'  : method_name    ,
                                    'method_kwargs': method_kwargs  ,
                                    'auth_key'     : auth_key       }}
        return requests.post(self.invoke_url, json=payload).json()


from osbot_aws.apis.Session import Session

from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self


class Cloud_Front(Kwargs_To_Self):

    def client(self):
        return Session().client('cloudfront',region_name='us-east-1')


    def distributions(self):
        response = self.client().list_distributions()       # todo: add support for when not all values are returned
        items = response.get('DistributionList', {}).get('Items', [])
        return items




# todo recreate with CLI util (see FastAPI recomendation)
# import os
# import boto3
# import json
# import shutil
#
# from osbot_aws.apis.Session import Session
#
#
# class Aws_Cli:
#     def __init__(self, **kwargs):
#         self.aws_lambda = Session().client('lambda', region_name = kwargs.get('region_name'))
#         self.aws_s3     = Session().client('s3'    , region_name = kwargs.get('region_name'))
#
#
#         #
#
#     # def lambda_delete_function      (self, name                                                                 ):
#     #     if self.lambda_function_exists(name):
#     #         self.aws_lambda.delete_function( FunctionName = name)
#     #     return self
#     #     #
#
#     # def lambda_functions            (self                                                                       ):
#     #     data = {}
#     #     for function in self.aws_lambda.list_functions().get('Functions'):
#     #         data[function['FunctionName']] = function
#     #     return data
#
#
#
#         #
#
#     # def lambda_function_info        (self, name                                                                 ):
#     #     return self.aws_lambda.get_function(FunctionName = name)
#
#     def lambda_invoke_function      (self, name, payload                                                        ):
#         response      = self.aws_lambda.invoke(FunctionName=name, Payload = json.dumps(payload) )
#
#         result_bytes  = response.get('Payload').read()
#         result_string = result_bytes.decode('utf-8')
#         result        = json.loads(result_string)
#         return result, response
#
#         #
#
#     # def lambda_invoke_function_async(self, name, payload                                                        ):
#     #     return  self.aws_lambda.invoke(FunctionName=name, Payload = json.dumps(payload), InvocationType='Event')
#
#     def lambda_update_function      (self, name, s3_bucket,s3_key                                               ):
#         return self.aws_lambda.update_function_code(FunctionName = name,
#                                                     S3Bucket     = s3_bucket ,
#                                                     S3Key        = s3_key    )
#
#         #
#
#     def s3_buckets                  (self                                                                       ):
#         data = {}
#         for bucket in self.aws_s3.list_buckets().get('Buckets'):
#             data[bucket['Name']] = bucket
#         return data
#
#         #
#
#     #def s3_add_notification        (self, s)



from osbot_aws.aws.bedrock.models.diffusion.Stability_Stable_Diffusion import Stability_Stable_Diffusion
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Png import save_png_base64_to_file


class Stability_Stable_Diffusion_XL_V0(Stability_Stable_Diffusion):
    model_id: str = 'stability.stable-diffusion-xl-v0'

    def body(self):
        return dict(text_prompts = self.text_prompts ,
                    cfg_scale    = self.cfg_scale    ,   # min: 0     max: 30     default: 10
                    steps        = self.steps        ,   # min: 10    max: 150    default: 30
                    seed         = self.seed         )





from osbot_aws.aws.iam.IAM_Role import IAM_Role

class IAM_Role_With_Policy(IAM_Role):

    def __init__(self, role_name=None):
        super().__init__(role_name)

    def policy_template_cloudwatch_allow_all(self):
        return {   "Version": "2012-10-17",
                   "Statement": [ {    "Effect": "Allow",
                                       "Action": [ "logs:CreateLogGroup"    ,
                                                   "logs:CreateLogStream"   ,
                                                   "logs:DescribeLogGroups" ,
                                                   "logs:DescribeLogStreams",
                                                   "logs:PutLogEvents"      ,
                                                   "logs:GetLogEvents"      ,
                                                   "logs:FilterLogEvents"  ],
                                       "Resource": "*" }]}

    def create_api_gateway__cloudwatch_allow_all(self, project_name=None):
        return self.create_cloudwatch_allow_all('apigateway.amazonaws.com', project_name)

    def create_cloudwatch_allow_all(self,service, project_name=None):
        policies = { "Cloud_Watch" : self.policy_template_cloudwatch_allow_all()}
        return self.create_for_service_with_policies(service, policies, project_name)


from unittest import TestCase

from botocore.client import BaseClient
from dotenv import load_dotenv

from osbot_aws.aws.boto3.Cache_Boto3_Requests import Cache_Boto3_Requests
from osbot_aws.testing.Pytest import skip_pytest___aws_pytest_user_name__is_not_set


class TestCase__Boto3_Cache(TestCase):
    cache : Cache_Boto3_Requests

    @classmethod
    def setUpClass(cls):
        skip_pytest___aws_pytest_user_name__is_not_set()
        load_dotenv()
        cls.cache = Cache_Boto3_Requests()
        cls.cache.patch_apply()

    @classmethod
    def tearDownClass(cls):
        cls.cache.patch_restore()
        assert BaseClient._make_api_call.__qualname__ == 'BaseClient._make_api_call'  # confirm the original function is still there



import sys


class Session_Bad_Credentials(Exception):

    def __init__(self, exception, print_error_message= True):
        error_code    = exception.get("Code")                               # todo: refactor into separate methods
        error_message = exception.get('Message')
        message = (f"\tCurrent Security credentials are invalid! \n\n"
                   f"\tPlease check: current environment values, "
                   f"profile name or value in your ~/.aws/credentials file"
                   f"\n\n"
                   f"\tBoto3 Error message: {error_code}  {error_message}")
        self.message             = message
        self.print_error_message = print_error_message
        if self.print_error_message:
            print()
            print("\n*********** OSBot_AWS ***********\n")
            print(f"{self.message}")
            print("\n*********** OSBot_AWS ***********\n")

        super().__init__(self.message)